# -*- coding: utf-8 -*-
"""
npxpy
Created on Thu Feb 29 11:49:17 2024

@author: Caghan Uenlueer
Neuromorphic Quantumphotonics
Heidelberg University
E-Mail:	caghan.uenlueer@kip.uni-heidelberg.de

This file is part of npxpy, which is licensed under the MIT License.
"""
from typing import Dict, Any, List, Union
from npxpy.nodes.node import Node
from npxpy.resources import Image


class CoarseAligner(Node):
    """
    Class for coarse alignment nodes.

    Attributes:
        residual_threshold (Union[float, int]): The residual threshold for alignment.
        alignment_anchors (List[Dict]): List of alignment anchors with label and position.
    """

    def __init__(
        self,
        name: str = "Coarse aligner",
        residual_threshold: Union[float, int] = 10.0,
    ):
        """
        Initialize the coarse aligner with a name and a residual threshold.

        Parameters:
            name (str): The name of the coarse aligner node.
            residual_threshold (Union[float, int]): The residual threshold for alignment. Must be greater than 0.

        Raises:
            ValueError: If residual_threshold is not greater than 0.
        """
        super().__init__("coarse_alignment", name)

        self._alignment_anchors = []
        self.residual_threshold = (
            residual_threshold  # Using setter for validation
        )

    @property
    def residual_threshold(self):
        """Get the residual threshold."""
        return self._residual_threshold

    @residual_threshold.setter
    def residual_threshold(self, value: Union[float, int]):
        """
        Set the residual threshold for alignment.

        Parameters:
            value (Union[float, int]): Residual threshold, must be greater than 0.

        Raises:
            ValueError: If residual_threshold is not greater than 0.
        """
        if not isinstance(value, (float, int)) or value <= 0:
            raise ValueError("residual_threshold must be a positive number.")
        self._residual_threshold = value

    @property
    def alignment_anchors(self):
        """Get the list of alignment anchors."""
        return self._alignment_anchors

    def add_coarse_anchor(self, position: List[Union[float, int]], label: str):
        """
        Add a single coarse anchor with a label and position.

        Parameters:
            label (str): The label for the anchor.
            position (List[Union[float, int]]): The position [x, y, z] for the anchor.

        Raises:
            ValueError: If position does not contain exactly three elements.
            TypeError: If any element in position is not a number.
        """
        if not isinstance(label, str):
            raise TypeError("label must be a string.")
        if len(position) != 3:
            raise ValueError("position must be a list of three elements.")
        if not all(isinstance(p, (float, int)) for p in position):
            raise TypeError("All position elements must be numbers.")

        self._alignment_anchors.append(
            {
                "label": label,
                "position": position,
            }
        )
        return self

    def set_coarse_anchors_at(
        self,
        positions: List[List[Union[float, int]]],
        labels: List[str] = None,
    ):
        """
        Create multiple coarse anchors at specified positions.

        Parameters:
            labels (List[str]): List of labels for the anchors.
            positions (List[List[Union[float, int]]]): List of positions for the anchors, each position is [x, y, z].

        Returns:
            self: The instance of the CoarseAligner class.

        Raises:
            ValueError: If the number of labels does not match the number of positions.
            TypeError: If any label is not a string or any position is not a list of numbers.
        """
        if labels is None:
            labels = [f"anchor_{i}" for i in range(len(positions))]
        if len(labels) != len(positions):
            raise ValueError(
                "The number of labels must match the number of positions."
            )

        for label in labels:
            if not isinstance(label, str):
                raise TypeError("All labels must be strings.")

        for position in positions:
            if len(position) != 3:
                raise ValueError(
                    "Each position must be a list of three elements."
                )
            if not all(isinstance(p, (float, int)) for p in position):
                raise TypeError("All position elements must be numbers.")

        for label, position in zip(labels, positions):
            self.add_coarse_anchor(position, label)

        return self

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the CoarseAligner object into a dictionary.

        Returns:
            Dict[str, Any]: Dictionary representation of the object.
        """
        node_dict = super().to_dict()  # Get basic attributes from Node
        node_dict.update(
            {
                "alignment_anchors": self.alignment_anchors,
                "residual_threshold": self.residual_threshold,
            }
        )
        return node_dict


class InterfaceAligner(Node):
    """
    A class representing an interface aligner node, responsible for managing
    alignment settings and grid patterns for various operations.

    Attributes:
        alignment_anchors (List[Dict]): Stores the measurement locations (or interface anchors) for interface alignment.
        count (List[int]): The number of grid points in [x, y] direction.
        size (List[float]): The size of the grid in [width, height].
        pattern (str): The pattern used for grid or custom alignment.
    """

    def __init__(
        self,
        name: str = "Interface aligner",
        signal_type: str = "auto",
        detector_type: str = "auto",
        measure_tilt: bool = False,
        area_measurement: bool = False,
        center_stage: bool = True,
        action_upon_failure: str = "abort",
        laser_power: float = 0.5,
        scan_area_res_factors: List[float] = [1.0, 1.0],
        scan_z_sample_distance: float = 0.1,
        scan_z_sample_count: int = 51,
    ):
        """
        Initializes an InterfaceAligner node with specified settings.

        Parameters:
            name (str): Name of the interface aligner. Defaults to "Interface aligner".
            signal_type (str): The type of signal. Can be 'auto', 'fluorescence', or 'reflection'. Defaults to 'auto'.
            detector_type (str): The type of detector. Can be 'auto', 'confocal', 'camera', or 'camera_legacy'. Defaults to 'auto'.
            measure_tilt (bool): Whether to measure tilt. Defaults to False.
            area_measurement (bool): Whether to measure the area. Defaults to False.
            center_stage (bool): Whether to center the stage. Defaults to True.
            action_upon_failure (str): Action upon failure, can be 'abort' or 'ignore'. Defaults to 'abort'.
            laser_power (float): The power of the laser. Must be a positive number. Defaults to 0.5.
            scan_area_res_factors (List[float]): Resolution factors for the scan area. Defaults to [1.0, 1.0].
            scan_z_sample_distance (float): Distance between samples in the z-direction. Defaults to 0.1.
            scan_z_sample_count (int): Number of samples in the z-direction. Must be greater than 0. Defaults to 51.

        Raises:
            ValueError: If any input is not valid (e.g., invalid types or constraints like negative values).
        """
        super().__init__("interface_alignment", name)

        # Use setters for validation
        self.signal_type = signal_type
        self.detector_type = detector_type
        self.measure_tilt = measure_tilt
        self.area_measurement = area_measurement
        self.center_stage = center_stage
        self.action_upon_failure = action_upon_failure
        self.laser_power = laser_power
        self.scan_area_res_factors = scan_area_res_factors
        self.scan_z_sample_distance = scan_z_sample_distance
        self.scan_z_sample_count = scan_z_sample_count

        self.alignment_anchors = []
        self.count = [5, 5]
        self.size = [200.0, 200.0]
        self.pattern = "Origin"

    # Setters with validation for various attributes

    @property
    def signal_type(self):
        return self._signal_type

    @signal_type.setter
    def signal_type(self, value: str):
        valid_types = ["auto", "fluorescence", "reflection"]
        if value not in valid_types:
            raise ValueError(f"signal_type must be one of {valid_types}.")
        self._signal_type = value

    @property
    def detector_type(self):
        return self._detector_type

    @detector_type.setter
    def detector_type(self, value: str):
        valid_detectors = ["auto", "confocal", "camera", "camera_legacy"]
        if value not in valid_detectors:
            raise ValueError(
                f"detector_type must be one of {valid_detectors}."
            )
        self._detector_type = value

    @property
    def measure_tilt(self):
        return self._measure_tilt

    @measure_tilt.setter
    def measure_tilt(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("measure_tilt must be a boolean.")
        self._measure_tilt = value

    @property
    def area_measurement(self):
        return self._area_measurement

    @area_measurement.setter
    def area_measurement(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("area_measurement must be a boolean.")
        self._area_measurement = value

    @property
    def center_stage(self):
        return self._center_stage

    @center_stage.setter
    def center_stage(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("center_stage must be a boolean.")
        self._center_stage = value

    @property
    def action_upon_failure(self):
        return self._action_upon_failure

    @action_upon_failure.setter
    def action_upon_failure(self, value: str):
        valid_actions = ["abort", "ignore"]
        if value not in valid_actions:
            raise ValueError(
                f"action_upon_failure must be one of {valid_actions}."
            )
        self._action_upon_failure = value

    @property
    def laser_power(self):
        return self._laser_power

    @laser_power.setter
    def laser_power(self, value: float):
        if not isinstance(value, (float, int)) or value <= 0:
            raise ValueError("laser_power must be a positive number.")
        self._laser_power = value

    @property
    def scan_area_res_factors(self):
        return self._scan_area_res_factors

    @scan_area_res_factors.setter
    def scan_area_res_factors(self, value: List[float]):
        if len(value) != 2 or not all(
            isinstance(f, (float, int)) for f in value
        ):
            raise TypeError(
                "scan_area_res_factors must be a list of two floats or ints."
            )
        self._scan_area_res_factors = value

    @property
    def scan_z_sample_distance(self):
        return self._scan_z_sample_distance

    @scan_z_sample_distance.setter
    def scan_z_sample_distance(self, value: float):
        if not isinstance(value, (float, int)):
            raise TypeError(
                "scan_z_sample_distance must be a float or an int."
            )
        self._scan_z_sample_distance = value

    @property
    def scan_z_sample_count(self):
        return self._scan_z_sample_count

    @scan_z_sample_count.setter
    def scan_z_sample_count(self, value: int):
        if not isinstance(value, int) or value < 1:
            raise ValueError(
                "scan_z_sample_count must be an integer greater than 0."
            )
        self._scan_z_sample_count = value

    @property
    def count(self):
        return self._count

    @count.setter
    def count(self, value: List[int]):
        if len(value) != 2 or not all(isinstance(c, int) for c in value):
            raise ValueError("count must be a list of two integers.")
        self._count = value

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, value: List[float]):
        if len(value) != 2 or not all(
            isinstance(s, (float, int)) for s in value
        ):
            raise ValueError("size must be a list of two numbers.")
        self._size = value

    @property
    def pattern(self):
        return self._pattern

    @pattern.setter
    def pattern(self, value: str):
        if value not in ["Grid", "Custom", "Origin"]:
            raise ValueError("pattern must be 'Grid', 'Custom', or 'Origin'.")
        self._pattern = value

    def set_grid(self, count: List[int], size: List[float]):
        """
        Sets the grid point count and grid size for alignment operations.

        Parameters:
            count (List[int]): Number of grid points in [x, y] direction. Must contain exactly two integers.
            size (List[float]): Size of the grid in [width, height]. Must contain exactly two numbers.

        Returns:
            self: The instance of the InterfaceAligner class.

        Raises:
            ValueError: If count or size does not contain exactly two elements.
            TypeError: If elements in count or size are not numbers.
        """
        self.count = count
        self.size = size
        self.pattern = "Grid"
        return self

    def add_interface_anchor(
        self,
        position: List[float],
        label: str,
        scan_area_size: List[float] = None,
    ):
        """
        Adds a custom interface anchor with a label, position, and optional scan area size.

        Parameters:
            label (str): The label for the anchor. Must be a string.
            position (List[float]): The position of the anchor [x, y]. Must contain exactly two numbers.
            scan_area_size (List[float], optional): The scan area size [width, height]. Defaults to [10.0, 10.0].

        Raises:
            ValueError: If position does not contain exactly two elements.
            TypeError: If label is not a string or elements in position or scan_area_size are not numbers.
        """
        if not isinstance(position, list) or len(position) != 2:
            try:
                position = list(position)
                position = position[:2]
                assert len(position) == 2
            except:
                raise ValueError("position must be a list of two elements.")
        if not all(isinstance(p, (float, int)) for p in position):
            try:
                position = [float(p) for p in position]
            except:
                raise TypeError("All position elements must be numbers.")
        if scan_area_size is None:
            scan_area_size = [10.0, 10.0]

        self.pattern = "Custom"
        self.alignment_anchors.append(
            {
                "label": label,
                "position": position,
                "scan_area_size": scan_area_size,
            }
        )
        return self

    def set_interface_anchors_at(
        self,
        positions: List[List[float]],
        labels: List[str] = None,
        scan_area_sizes: List[List[float]] = None,
    ):
        """
        Creates multiple custom interface anchors at specified positions.

        Parameters:
            labels (List[str]): List of labels for the measurement locations.
            positions (List[List[float]]): List of positions for the measurement locations, each position is [x, y].
            scan_area_sizes (List[List[float]], optional): List of scan area sizes for the measurement locations,
                                                           each scan area size is [width, height]. Defaults to [10.0, 10.0]
                                                           for each anchor.

        Returns:
            self: The instance of the InterfaceAligner class.

        Raises:
            ValueError: If the number of labels does not match the number of positions.
            TypeError: If elements in labels, positions, or scan_area_sizes are not of the correct types.
        """
        if scan_area_sizes is None:
            scan_area_sizes = [[10.0, 10.0]] * len(positions)
        if labels is None:
            labels = [f"anchor_{i}" for i in range(len(positions))]
        for label, position, scan_area_size in zip(
            labels, positions, scan_area_sizes
        ):
            self.add_interface_anchor(position, label, scan_area_size)

        return self

    def to_dict(self) -> Dict:
        """
        Converts the current state of the object into a dictionary representation.

        Returns:
            dict: Dictionary representation of the current state of the object, including
                  alignment anchors, grid settings, and signal properties.
        """
        node_dict = super().to_dict()
        if self.signal_type == "auto" or self.detector_type == "camera_legacy":
            node_dict["interface_finder_type"] = self.signal_type
        else:
            node_dict["interface_finder_type"] = (
                f"{self.signal_type}_{self.detector_type}"
            )
        node_dict["properties"] = {
            "signal_type": self.signal_type,
            "detector_type": self.detector_type,
        }
        node_dict["alignment_anchors"] = self.alignment_anchors
        node_dict["grid_point_count"] = self.count
        node_dict["grid_size"] = self.size
        node_dict["pattern"] = self.pattern
        node_dict["measure_tilt"] = self.measure_tilt
        node_dict["area_measurement"] = self.area_measurement
        node_dict["center_stage"] = self.center_stage
        node_dict["action_upon_failure"] = self.action_upon_failure
        node_dict["laser_power"] = self.laser_power
        node_dict["scan_area_res_factors"] = self.scan_area_res_factors
        node_dict["scan_z_sample_distance"] = self.scan_z_sample_distance
        node_dict["scan_z_sample_count"] = self.scan_z_sample_count
        return node_dict


class FiberAligner(Node):
    def __init__(
        self,
        name: str = "Fiber aligner",
        fiber_radius: Union[float, int] = 63.5,
        center_stage: bool = True,
        action_upon_failure: str = "abort",
        illumination_name: str = "process_led_1",
        core_signal_lower_threshold: Union[float, int] = 0.05,
        core_signal_range: List[Union[float, int]] = [0.1, 0.9],
        detection_margin: Union[float, int] = 6.35,
    ):
        """
        Initialize the fiber aligner with specified parameters.

        Parameters:
            name (str): Name of the fiber aligner.
            fiber_radius (Union[float, int]): Radius of the fiber.
            center_stage (bool): Whether to center the stage.
            action_upon_failure (str): Action upon failure ('abort' or 'ignore').
            illumination_name (str): Name of the illumination source.
            core_signal_lower_threshold (Union[float, int]): Lower threshold for
                the core signal.
            core_signal_range (List[Union[float, int]]): Range for the core
                signal [min, max].
            detection_margin (Union[float, int]): Detection margin.
        """
        super().__init__(
            node_type="fiber_core_alignment",
            name=name,
        )

        # Use setters to handle attributes
        self.fiber_radius = fiber_radius
        self.center_stage = center_stage
        self.action_upon_failure = action_upon_failure
        self.illumination_name = illumination_name
        self.core_signal_lower_threshold = core_signal_lower_threshold
        self.core_signal_range = core_signal_range
        self.detection_margin = detection_margin

        # Default values using setters for consistency
        self.detect_light_direction = False
        self.z_scan_range = [10, 100]
        self.z_scan_range_sample_count = 1
        self.z_scan_range_scan_count = 1

    # Proper setters with validation
    @property
    def fiber_radius(self) -> Union[float, int]:
        return self._fiber_radius

    @fiber_radius.setter
    def fiber_radius(self, value: Union[float, int]):
        if not isinstance(value, (float, int)) or value <= 0:
            raise ValueError("fiber_radius must be a positive number.")
        self._fiber_radius = value

    @property
    def center_stage(self) -> bool:
        return self._center_stage

    @center_stage.setter
    def center_stage(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("center_stage must be a boolean.")
        self._center_stage = value

    @property
    def action_upon_failure(self) -> str:
        return self._action_upon_failure

    @action_upon_failure.setter
    def action_upon_failure(self, value: str):
        if value not in ["abort", "ignore"]:
            raise ValueError(
                "action_upon_failure must be 'abort' or 'ignore'."
            )
        self._action_upon_failure = value

    @property
    def illumination_name(self) -> str:
        return self._illumination_name

    @illumination_name.setter
    def illumination_name(self, value: str):
        if not isinstance(value, str):
            raise TypeError("illumination_name must be a string.")
        self._illumination_name = value

    @property
    def core_signal_lower_threshold(self) -> Union[float, int]:
        return self._core_signal_lower_threshold

    @core_signal_lower_threshold.setter
    def core_signal_lower_threshold(self, value: Union[float, int]):
        if not isinstance(value, (float, int)):
            raise TypeError(
                "core_signal_lower_threshold must be a float or an int."
            )
        self._core_signal_lower_threshold = value

    @property
    def core_signal_range(self) -> List[Union[float, int]]:
        return self._core_signal_range

    @core_signal_range.setter
    def core_signal_range(self, value: List[Union[float, int]]):
        if not isinstance(value, list) or len(value) != 2:
            raise ValueError(
                "core_signal_range must be a list of two elements."
            )
        if not all(isinstance(val, (float, int)) for val in value):
            raise TypeError(
                "All elements in core_signal_range must be numbers."
            )
        self._core_signal_range = value

    @property
    def detection_margin(self) -> Union[float, int]:
        return self._detection_margin

    @detection_margin.setter
    def detection_margin(self, value: Union[float, int]):
        if not isinstance(value, (float, int)) or value <= 0:
            raise ValueError("detection_margin must be a positive number.")
        self._detection_margin = value

    @property
    def z_scan_range(self) -> List[Union[float, int]]:
        return self._z_scan_range

    @z_scan_range.setter
    def z_scan_range(self, value: List[Union[float, int]]):
        if (
            not isinstance(value, list)
            or len(value) != 2
            or value[1] <= value[0]
        ):
            raise ValueError(
                "z_scan_range must be a list of two elements where the "
                "second element is greater than the first."
            )
        self._z_scan_range = value

    @property
    def z_scan_range_sample_count(self) -> int:
        return self._z_scan_range_sample_count

    @z_scan_range_sample_count.setter
    def z_scan_range_sample_count(self, value: int):
        if not isinstance(value, int) or value <= 0:
            raise ValueError(
                "z_scan_range_sample_count must be a positive integer."
            )
        self._z_scan_range_sample_count = value

    @property
    def z_scan_range_scan_count(self) -> int:
        return self._z_scan_range_scan_count

    @z_scan_range_scan_count.setter
    def z_scan_range_scan_count(self, value: int):
        if not isinstance(value, int) or value <= 0:
            raise ValueError(
                "z_scan_range_scan_count must be a positive integer."
            )
        self._z_scan_range_scan_count = value

    def measure_tilt(
        self,
        z_scan_range: List[Union[float, int]] = [10, 100],
        z_scan_range_sample_count: int = 3,
        z_scan_range_scan_count: int = 1,
    ):
        """
        Measures tilt by setting scan range parameters.

        Parameters:
            z_scan_range (List[Union[float, int]]): Range for the z-scan.
            z_scan_range_sample_count (int): Number of samples in the z-scan.
            z_scan_range_scan_count (int): Number of scans in the z-scan.

        Returns:
            self: The instance of the FiberAligner class.

        """
        self.z_scan_range = z_scan_range
        self.z_scan_range_sample_count = z_scan_range_sample_count
        self.z_scan_range_scan_count = z_scan_range_scan_count
        self.detect_light_direction = True

        return self

    def to_dict(self) -> Dict:
        """
        Converts the current state of the object into a dictionary.

        Returns:
            dict: Dictionary representation of the current state of the object.
        """
        node_dict = super().to_dict()
        # Add custom attributes not covered by super().__init__()
        node_dict.update(
            {
                "fiber_radius": self.fiber_radius,
                "center_stage": self.center_stage,
                "action_upon_failure": self.action_upon_failure,
                "illumination_name": self.illumination_name,
                "core_signal_lower_threshold": self.core_signal_lower_threshold,
                "core_signal_range": self.core_signal_range,
                "core_position_offset_tolerance": self.detection_margin,
                "detect_light_direction": self.detect_light_direction,
                "z_scan_range": self.z_scan_range,
                "z_scan_range_sample_count": self.z_scan_range_sample_count,
                "z_scan_range_scan_count": self.z_scan_range_scan_count,
            }
        )
        return node_dict


class MarkerAligner(Node):
    """
    Marker aligner class.

    Attributes:
        image (Resources): Image object that the marker gets assigned.
        name (str): Name of the marker aligner.
        marker_size (List[float]): Size of markers in micrometers. Marker size must be greater than 0.
        center_stage (bool): Centers stage if true.
        action_upon_failure (str): 'abort' or 'ignore' at failure (not yet implemented!).
        laser_power (float): Laser power in mW.
        scan_area_size (List[float]): Scan area size in micrometers.
        scan_area_res_factors (List[float]): Resolution factors in scanned area.
        detection_margin (float): Additional margin around marker imaging field in micrometers.
        correlation_threshold (float): Correlation threshold below which abort is triggered in percent.
        residual_threshold (float): Residual threshold of marker image.
        max_outliers (int): Maximum amount of markers that are allowed to be outliers.
        orthonormalize (bool): Whether to orthonormalize or not.
        z_scan_sample_count (int): Number of z samples to be taken.
        z_scan_sample_distance (float): Sampling distance in micrometers for z samples to be apart from each other.
        z_scan_sample_mode (str): "correlation" or "intensity" for scan_z_sample_mode.
        measure_z (bool): Whether to measure z or not.
    """

    def __init__(
        self,
        image: Image,
        name: str = "Marker aligner",
        marker_size: List[float] = [5.0, 5.0],
        center_stage: bool = True,
        action_upon_failure: str = "abort",
        laser_power: float = 0.5,
        scan_area_size: List[float] = [10.0, 10.0],
        scan_area_res_factors: List[float] = [2.0, 2.0],
        detection_margin: float = 5.0,
        correlation_threshold: float = 60.0,
        residual_threshold: float = 0.5,
        max_outliers: int = 0,
        orthonormalize: bool = True,
        z_scan_sample_count: int = 1,
        z_scan_sample_distance: float = 0.5,
        z_scan_sample_mode: str = "correlation",
        measure_z: bool = False,
    ):
        """
        Initializes the MarkerAligner with the provided parameters.
        """
        super().__init__(node_type="marker_alignment", name=name)

        # Set attributes via setters
        self.image = image
        self.marker_size = marker_size
        self.center_stage = center_stage
        self.action_upon_failure = action_upon_failure
        self.laser_power = laser_power
        self.scan_area_size = scan_area_size
        self.scan_area_res_factors = scan_area_res_factors
        self.detection_margin = detection_margin
        self.correlation_threshold = correlation_threshold
        self.residual_threshold = residual_threshold
        self.max_outliers = max_outliers
        self.orthonormalize = orthonormalize
        self.z_scan_sample_count = z_scan_sample_count
        self.z_scan_sample_distance = z_scan_sample_distance
        self.z_scan_sample_mode = z_scan_sample_mode
        self.measure_z = measure_z

        self.alignment_anchors = []

    # Property setters with validation
    @property
    def image(self) -> Image:
        return self._image

    @image.setter
    def image(self, value: Image):
        if not isinstance(value, Image):
            raise TypeError("image must be an instance of Image class.")
        self._image = value

    @property
    def marker_size(self) -> List[float]:
        return self._marker_size

    @marker_size.setter
    def marker_size(self, value: List[float]):
        if (
            not isinstance(value, list)
            or len(value) != 2
            or not all(isinstance(val, (float, int)) for val in value)
        ):
            raise TypeError(
                "marker_size must be a list of two positive numbers."
            )
        if value[0] <= 0 or value[1] <= 0:
            raise ValueError("marker_size must be greater than 0.")
        self._marker_size = value

    @property
    def center_stage(self) -> bool:
        return self._center_stage

    @center_stage.setter
    def center_stage(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("center_stage must be a boolean.")
        self._center_stage = value

    @property
    def action_upon_failure(self) -> str:
        return self._action_upon_failure

    @action_upon_failure.setter
    def action_upon_failure(self, value: str):
        if value not in ["abort", "ignore"]:
            raise ValueError(
                "action_upon_failure must be 'abort' or 'ignore'."
            )
        self._action_upon_failure = value

    @property
    def laser_power(self) -> float:
        return self._laser_power

    @laser_power.setter
    def laser_power(self, value: float):
        if not isinstance(value, (float, int)) or value < 0:
            raise ValueError("laser_power must be a non-negative number.")
        self._laser_power = value

    @property
    def scan_area_size(self) -> List[float]:
        return self._scan_area_size

    @scan_area_size.setter
    def scan_area_size(self, value: List[float]):
        if (
            not isinstance(value, list)
            or len(value) != 2
            or not all(isinstance(val, (float, int)) for val in value)
        ):
            raise TypeError("scan_area_size must be a list of two numbers.")
        self._scan_area_size = value

    @property
    def scan_area_res_factors(self) -> List[float]:
        return self._scan_area_res_factors

    @scan_area_res_factors.setter
    def scan_area_res_factors(self, value: List[float]):
        if (
            not isinstance(value, list)
            or len(value) != 2
            or not all(isinstance(val, (float, int)) for val in value)
        ):
            raise TypeError(
                "scan_area_res_factors must be a list of two numbers."
            )
        self._scan_area_res_factors = value

    @property
    def detection_margin(self) -> float:
        return self._detection_margin

    @detection_margin.setter
    def detection_margin(self, value: float):
        if not isinstance(value, (float, int)) or value < 0:
            raise ValueError("detection_margin must be a non-negative number.")
        self._detection_margin = value

    @property
    def correlation_threshold(self) -> float:
        return self._correlation_threshold

    @correlation_threshold.setter
    def correlation_threshold(self, value: float):
        if not isinstance(value, (float, int)) or not (0 <= value <= 100):
            raise ValueError(
                "correlation_threshold must be between 0 and 100."
            )
        self._correlation_threshold = value

    @property
    def residual_threshold(self) -> float:
        return self._residual_threshold

    @residual_threshold.setter
    def residual_threshold(self, value: float):
        if not isinstance(value, (float, int)) or value < 0:
            raise ValueError(
                "residual_threshold must be a non-negative number."
            )
        self._residual_threshold = value

    @property
    def max_outliers(self) -> int:
        return self._max_outliers

    @max_outliers.setter
    def max_outliers(self, value: int):
        if not isinstance(value, int) or value < 0:
            raise ValueError("max_outliers must be a non-negative integer.")
        self._max_outliers = value

    @property
    def orthonormalize(self) -> bool:
        return self._orthonormalize

    @orthonormalize.setter
    def orthonormalize(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("orthonormalize must be a boolean.")
        self._orthonormalize = value

    @property
    def z_scan_sample_count(self) -> int:
        return self._z_scan_sample_count

    @z_scan_sample_count.setter
    def z_scan_sample_count(self, value: int):
        if not isinstance(value, int) or value < 1:
            raise ValueError("z_scan_sample_count must be at least 1.")
        self._z_scan_sample_count = value

    @property
    def z_scan_sample_distance(self) -> float:
        return self._z_scan_sample_distance

    @z_scan_sample_distance.setter
    def z_scan_sample_distance(self, value: float):
        if not isinstance(value, (float, int)) or value <= 0:
            raise ValueError(
                "z_scan_sample_distance must be a positive number."
            )
        self._z_scan_sample_distance = value

    @property
    def z_scan_sample_mode(self) -> str:
        return self._z_scan_sample_mode

    @z_scan_sample_mode.setter
    def z_scan_sample_mode(self, value: str):
        if value not in ["correlation", "intensity"]:
            raise ValueError(
                'z_scan_sample_mode must be either "correlation" or "intensity".'
            )
        self._z_scan_sample_mode = value

    @property
    def measure_z(self) -> bool:
        return self._measure_z

    @measure_z.setter
    def measure_z(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("measure_z must be a boolean.")
        self._measure_z = value

    def add_marker(
        self, position: List[float], orientation: float, label: str
    ):
        """
        Adds a marker to the alignment anchors.
        """
        if not isinstance(label, str):
            raise TypeError("label must be a string.")
        if not isinstance(orientation, (float, int)):
            try:
                float(orientation)
            except:
                raise TypeError("orientation must be a float or an int.")
        if (
            not isinstance(position, list)
            or len(position) != 3
            or not all(isinstance(val, (float, int)) for val in position)
        ):
            raise TypeError("position must be a list of three numbers.")

        self.alignment_anchors.append(
            {"label": label, "position": position, "rotation": orientation}
        )
        return self

    def set_markers_at(
        self,
        positions: List[List[float]],
        orientations: List[float] = None,
        labels: List[str] = None,
    ):
        """
        Creates multiple markers at specified positions with given orientations.
        """
        if labels is None:
            labels = [f"marker_{i}" for i in range(len(positions))]
        if orientations is None:
            orientations = [0 for i in range(len(positions))]
        if len(labels) != len(positions) or len(labels) != len(orientations):
            raise ValueError(
                "The number of labels, positions, and orientations must match."
            )

        for label in labels:
            if not isinstance(label, str):
                raise TypeError("All labels must be strings.")

        for position in positions:
            if (
                not isinstance(position, list)
                or len(position) != 3
                or not all(isinstance(val, (float, int)) for val in position)
            ):
                raise TypeError(
                    "All positions must be lists of three numbers."
                )

        for label, orientation, position in zip(
            labels, orientations, positions
        ):
            self.add_marker(position, orientation, label)
        return self

    def to_dict(self) -> Dict:
        """
        Converts the current state of the object into a dictionary representation.
        """
        node_dict = super().to_dict()
        node_dict.update(
            {
                "marker": {"image": self.image.id, "size": self.marker_size},
                "center_stage": self.center_stage,
                "action_upon_failure": self.action_upon_failure,
                "laser_power": self.laser_power,
                "scan_area_size": self.scan_area_size,
                "scan_area_res_factors": self.scan_area_res_factors,
                "detection_margin": self.detection_margin,
                "correlation_threshold": self.correlation_threshold,
                "residual_threshold": self.residual_threshold,
                "max_outliers": self.max_outliers,
                "orthonormalize": self.orthonormalize,
                "z_scan_sample_count": self.z_scan_sample_count,
                "z_scan_sample_distance": self.z_scan_sample_distance,
                "z_scan_optimization_mode": self.z_scan_sample_mode,
                "measure_z": self.measure_z,
                "alignment_anchors": self.alignment_anchors,
            }
        )
        return node_dict


class EdgeAligner(Node):
    """
    A class to represent an edge aligner with various attributes and methods for managing edge alignment.

    Attributes:
        alignment_anchors (List[Dict[str, Any]]): List of alignment anchors.
    """

    def __init__(
        self,
        name: str = "Edge aligner",
        edge_location: List[float] = [0.0, 0.0],
        edge_orientation: float = 0.0,
        center_stage: bool = True,
        action_upon_failure: str = "abort",
        laser_power: Union[float, int] = 0.5,
        scan_area_res_factors: List[float] = [1.0, 1.0],
        scan_z_sample_distance: Union[float, int] = 0.1,
        scan_z_sample_count: int = 51,
        outlier_threshold: float = 10.0,
    ):
        """
        Initialize the edge aligner with the specified parameters.
        """
        super().__init__(node_type="edge_alignment", name=name)

        # Set attributes using setters
        self.edge_location = edge_location
        self.edge_orientation = edge_orientation
        self.center_stage = center_stage
        self.action_upon_failure = action_upon_failure
        self.laser_power = laser_power
        self.scan_area_res_factors = scan_area_res_factors
        self.scan_z_sample_distance = scan_z_sample_distance
        self.scan_z_sample_count = scan_z_sample_count
        self.outlier_threshold = outlier_threshold

        self.alignment_anchors = []

    # Property setters with validation
    @property
    def edge_location(self) -> List[float]:
        return self._edge_location

    @edge_location.setter
    def edge_location(self, value: List[float]):
        if (
            not isinstance(value, list)
            or len(value) != 2
            or not all(isinstance(val, (float, int)) for val in value)
        ):
            raise TypeError("edge_location must be a list of two numbers.")
        self._edge_location = value

    @property
    def edge_orientation(self) -> float:
        return self._edge_orientation

    @edge_orientation.setter
    def edge_orientation(self, value: float):
        if not isinstance(value, (float, int)):
            raise TypeError("edge_orientation must be a float or an int.")
        self._edge_orientation = value

    @property
    def center_stage(self) -> bool:
        return self._center_stage

    @center_stage.setter
    def center_stage(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("center_stage must be a boolean.")
        self._center_stage = value

    @property
    def action_upon_failure(self) -> str:
        return self._action_upon_failure

    @action_upon_failure.setter
    def action_upon_failure(self, value: str):
        if value not in ["abort", "ignore"]:
            raise ValueError(
                "action_upon_failure must be 'abort' or 'ignore'."
            )
        self._action_upon_failure = value

    @property
    def laser_power(self) -> Union[float, int]:
        return self._laser_power

    @laser_power.setter
    def laser_power(self, value: Union[float, int]):
        if not isinstance(value, (float, int)) or value < 0:
            raise ValueError("laser_power must be a non-negative number.")
        self._laser_power = value

    @property
    def scan_area_res_factors(self) -> List[float]:
        return self._scan_area_res_factors

    @scan_area_res_factors.setter
    def scan_area_res_factors(self, value: List[float]):
        if (
            not isinstance(value, list)
            or len(value) != 2
            or not all(isinstance(val, (float, int)) for val in value)
        ):
            raise TypeError(
                "scan_area_res_factors must be a list of two numbers greater than zero."
            )
        if not all(factor > 0 for factor in value):
            raise ValueError(
                "All elements in scan_area_res_factors must be greater than 0."
            )
        self._scan_area_res_factors = value

    @property
    def scan_z_sample_distance(self) -> Union[float, int]:
        return self._scan_z_sample_distance

    @scan_z_sample_distance.setter
    def scan_z_sample_distance(self, value: Union[float, int]):
        if not isinstance(value, (float, int)) or value <= 0:
            raise ValueError(
                "scan_z_sample_distance must be a positive number."
            )
        self._scan_z_sample_distance = value

    @property
    def scan_z_sample_count(self) -> int:
        return self._scan_z_sample_count

    @scan_z_sample_count.setter
    def scan_z_sample_count(self, value: int):
        if not isinstance(value, int) or value < 1:
            raise ValueError(
                "scan_z_sample_count must be an integer greater than zero."
            )
        self._scan_z_sample_count = value

    @property
    def outlier_threshold(self) -> float:
        return self._outlier_threshold

    @outlier_threshold.setter
    def outlier_threshold(self, value: float):
        if not isinstance(value, (float, int)) or not (0 <= value <= 100):
            raise ValueError(
                "outlier_threshold must be a number between 0 and 100."
            )
        self._outlier_threshold = value

    def add_measurement(
        self,
        offset: Union[float, int],
        scan_area_size: List[Union[float, int]],
        label: str,
    ):
        """
        Add a measurement with a label, offset, and scan area size.
        """
        if not isinstance(label, str):
            raise TypeError("label must be a string.")
        if not isinstance(offset, (float, int)):
            raise TypeError("offset must be a float or an int.")
        if (
            not isinstance(scan_area_size, list)
            or len(scan_area_size) != 2
            or not all(isinstance(val, (float, int)) for val in scan_area_size)
        ):
            raise TypeError("scan_area_size must be a list of two numbers.")
        if scan_area_size[0] <= 0:
            raise ValueError(
                "The width (X) in scan_area_size must be greater than 0."
            )
        if scan_area_size[1] < 0:
            raise ValueError(
                "The height (Y) in scan_area_size must be greater than or equal to 0."
            )

        self.alignment_anchors.append(
            {
                "label": label,
                "offset": offset,
                "scan_area_size": scan_area_size,
            }
        )
        return self

    def set_measurements_at(
        self,
        offsets: List[Union[float, int]],
        scan_area_sizes: List[List[Union[float, int]]] = None,
        labels: List[str] = None,
    ):
        """
        Set multiple measurements at specified positions.
        """
        if scan_area_sizes is None:
            scan_area_sizes = [[50.0, 10.0]] * len(offsets)
        if labels is None:
            labels = [f"marker_{i}" for i in range(len(offsets))]
        if len(labels) != len(scan_area_sizes) or len(labels) != len(offsets):
            raise ValueError(
                "The number of labels, offsets, and scan_area_sizes must match."
            )

        for label in labels:
            if not isinstance(label, str):
                raise TypeError("All labels must be strings.")

        for offset in offsets:
            if not isinstance(offset, (float, int)):
                raise TypeError("All offsets must be float or int.")

        for scan_area_size in scan_area_sizes:
            if (
                not isinstance(scan_area_size, list)
                or len(scan_area_size) != 2
                or not all(
                    isinstance(val, (float, int)) for val in scan_area_size
                )
            ):
                raise TypeError(
                    "All scan_area_sizes must be lists of two numbers."
                )
            if scan_area_size[0] <= 0:
                raise ValueError(
                    "The width (X) in scan_area_size must be greater than 0."
                )
            if scan_area_size[1] < 0:
                raise ValueError(
                    "The height (Y) in scan_area_size must be greater than or equal to 0."
                )

        for label, offset, scan_area_size in zip(
            labels, offsets, scan_area_sizes
        ):
            self.add_measurement(offset, scan_area_size, label)
        return self

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the current state of the object into a dictionary representation.
        """
        node_dict = super().to_dict()
        node_dict.update(
            {
                "xy_position_local_cos": self.edge_location,
                "z_rotation_local_cos": self.edge_orientation,
                "center_stage": self.center_stage,
                "action_upon_failure": self.action_upon_failure,
                "laser_power": self.laser_power,
                "scan_area_res_factors": self.scan_area_res_factors,
                "scan_z_sample_distance": self.scan_z_sample_distance,
                "scan_z_sample_count": self.scan_z_sample_count,
                "outlier_threshold": self.outlier_threshold,
                "alignment_anchors": self.alignment_anchors,
            }
        )
        return node_dict



# -*- coding: utf-8 -*-
"""
npxpy
Created on Thu Feb 29 11:49:17 2024

@author: Caghan Uenlueer
Neuromorphic Quantumphotonics
Heidelberg University
E-Mail:	caghan.uenlueer@kip.uni-heidelberg.de

This file is part of npxpy, which is licensed under the MIT License.
"""
from typing import Dict, List
from npxpy.nodes.node import Node


class _space_ops_space(Node):
    def position_at(
        self,
        position: List[float] = [0.0, 0.0, 0.0],
        rotation: List[float] = None,
    ):
        """
        Set the position and rotation of the scene/group/array.

        Parameters:
            position (List[float]): The new position [x, y, z].
            rotation (List[float]): The new rotation [psi, theta, phi].

        Returns:
            Acene/Group/Array: The updated object.
        """
        if rotation is not None:
            self.position = position
            self.rotation = rotation
        else:
            self.position = position
        return self

    def translate(
        self,
        translation: List[float] = [
            0.0,
            0.0,
            0.0,
        ],
    ):
        """
        Translate the current position by the specified values.

        Parameters:
            translation (List[float]): The translation values [dx, dy, dz].

        Raises:
            ValueError: If translation does not have exactly 3 elements.

        Returns:
            Acene/Group/Array: The updated object.
        """
        if (
            len(translation) != 3
            or not all(isinstance(t, (int, float)) for t in translation)
            or not isinstance(translation, list)
        ):
            raise ValueError(
                "Translation must be a list of three numeric elements."
            )
        self.position = [p + t for p, t in zip(self.position, translation)]
        return self

    def rotate(
        self,
        rotation: List[float] = [
            0.0,
            0.0,
            0.0,
        ],
    ):
        """
        Rotate the scene/group/array by specified angles.

        Parameters:
            rotation (List[float]): The rotation values [d_psi, d_theta, d_phi].

        Raises:
            ValueError: If rotation does not have exactly 3 elements.

        Returns:
            Acene/Group/Array: The updated object.
        """
        if len(rotation) != 3 or not all(
            isinstance(r, (int, float)) for r in rotation
        ):
            raise ValueError(
                "Rotation must be a list of three numeric elements."
            )
        self.rotation = [
            (r + delta) % 360 for r, delta in zip(self.rotation, rotation)
        ]
        return self


class Scene(_space_ops_space):
    """
    Class representing a scene node.

    Attributes:
        position (List[float]): Position of the scene [x, y, z].
        rotation (List[float]): Rotation of the scene [psi, theta, phi].
        writing_direction_upward (bool): Writing direction of the scene.
    """

    def __init__(
        self,
        name: str = "Scene",
        position: List[float] = [0, 0, 0],
        rotation: List[float] = [0.0, 0.0, 0.0],
        writing_direction_upward: bool = True,
    ):
        """
        Initialize a Scene node.
        """
        super().__init__("scene", name)
        self.position = position
        self.rotation = rotation
        self._writing_direction_upward = None
        self.writing_direction_upward = (
            writing_direction_upward  # Using setter
        )

    # Setters for position and rotation
    @property
    def position(self) -> List[float]:
        return self._position

    @position.setter
    def position(self, value: List[float]):
        if not isinstance(value, list) or len(value) != 3:
            raise ValueError(
                "Position must be a list of three numeric values."
            )
        self._position = value

    @property
    def rotation(self) -> List[float]:
        return self._rotation

    @rotation.setter
    def rotation(self, value: List[float]):
        if not isinstance(value, list) or len(value) != 3:
            raise ValueError(
                "Rotation must be a list of three numeric values."
            )
        self._rotation = value

    # Getter and setter for writing direction
    @property
    def writing_direction_upward(self):
        return self._writing_direction_upward

    @writing_direction_upward.setter
    def writing_direction_upward(self, value: bool):
        if not isinstance(value, bool):
            raise ValueError("writing_direction_upward must be a boolean.")
        self._writing_direction_upward = value

    def to_dict(self) -> Dict:
        """
        Convert the Scene object into a dictionary.
        """
        node_dict = super().to_dict()
        node_dict["position"] = self.position
        node_dict["rotation"] = self.rotation
        node_dict["writing_direction_upward"] = self.writing_direction_upward
        return node_dict


class Group(_space_ops_space):
    """
    Class representing a group node.

    Attributes:
        position (List[float]): Position of the group [x, y, z].
        rotation (List[float]): Rotation of the group [psi, theta, phi].
    """

    def __init__(
        self,
        name: str = "Group",
        position: List[float] = [0, 0, 0],
        rotation: List[float] = [0.0, 0.0, 0.0],
    ):
        """
        Initialize a Group node.
        """
        super().__init__("group", name)
        self.position = position
        self.rotation = rotation

    # Setters for position and rotation
    @property
    def position(self) -> List[float]:
        return self._position

    @position.setter
    def position(self, value: List[float]):
        if not isinstance(value, list) or len(value) != 3:
            raise ValueError(
                "Position must be a list of three numeric values."
            )
        self._position = value

    @property
    def rotation(self) -> List[float]:
        return self._rotation

    @rotation.setter
    def rotation(self, value: List[float]):
        if not isinstance(value, list) or len(value) != 3:
            raise ValueError(
                "Rotation must be a list of three numeric values."
            )
        self._rotation = value

    def to_dict(self) -> Dict:
        """
        Convert the Group object into a dictionary.
        """
        node_dict = super().to_dict()
        node_dict["position"] = self.position
        node_dict["rotation"] = self.rotation
        return node_dict


class Array(_space_ops_space):
    """
    Class representing an array node with additional attributes.

    Attributes:
        position (List[float]): Position of the array [x, y, z].
        rotation (List[float]): Rotation of the array [psi, theta, phi].
        count (List[int]): Number of grid points in [x, y] direction.
        spacing (List[float]): Spacing of the grid in [width, height].
        order (str): Order of the array ('Lexical' or 'Meander').
        shape (str): Shape of the array ('Rectangular' or 'Round').
    """

    def __init__(
        self,
        name: str = "Array",
        position: List[float] = [0, 0, 0],
        rotation: List[float] = [0.0, 0.0, 0.0],
        count: List[int] = [5, 5],
        spacing: List[float] = [100.0, 100.0],
        order: str = "Lexical",
        shape: str = "Rectangular",
    ):
        """
        Initialize an Array node.
        """
        super().__init__("array", name)
        self.position = position
        self.rotation = rotation
        self.count = count
        self.spacing = spacing
        self.order = order
        self.shape = shape

    # Setters for position and rotation
    @property
    def position(self) -> List[float]:
        return self._position

    @position.setter
    def position(self, value: List[float]):
        if not isinstance(value, list) or len(value) != 3:
            raise ValueError(
                "Position must be a list of three numeric values."
            )
        self._position = value

    @property
    def rotation(self) -> List[float]:
        return self._rotation

    @rotation.setter
    def rotation(self, value: List[float]):
        if not isinstance(value, list) or len(value) != 3:
            raise ValueError(
                "Rotation must be a list of three numeric values."
            )
        self._rotation = value

    # Setters for other attributes
    @property
    def count(self):
        return self._count

    @property
    def count(self):
        return self._count

    @count.setter
    def count(self, value: List[int]):
        if len(value) != 2 or not all(
            isinstance(c, int) and c > 0 for c in value
        ):
            raise ValueError(
                "Count must be a list of exactly two integers greater than zero."
            )
        self._count = value

    @property
    def spacing(self):
        return self._spacing

    @spacing.setter
    def spacing(self, value: List[float]):
        if len(value) != 2 or not all(
            isinstance(s, (int, float)) for s in value
        ):
            raise ValueError(
                "Spacing must be a list of exactly two numeric values."
            )
        self._spacing = value

    @property
    def order(self):
        return self._order

    @order.setter
    def order(self, value: str):
        if value not in ["Lexical", "Meander"]:
            raise ValueError("order must be either 'Lexical' or 'Meander'.")
        self._order = value

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, value: str):
        if value not in ["Rectangular", "Round"]:
            raise ValueError("shape must be either 'Rectangular' or 'Round'.")
        self._shape = value

    def set_grid(
        self, count: List[int] = count, spacing: List[float] = spacing
    ):
        """
        Set the count and spacing of the array grid.

        Parameters:
            count (List[int]): The new grid point count.
            spacing (List[float]): The new grid spacing.

        Returns:
            Array: The updated Array object.
        """
        self.count = count
        self.spacing = spacing
        return self

    def to_dict(self) -> Dict:
        """
        Convert the Array object into a dictionary.
        """
        node_dict = super().to_dict()
        node_dict["position"] = self.position
        node_dict["rotation"] = self.rotation
        node_dict["count"] = self.count
        node_dict["spacing"] = self.spacing
        node_dict["order"] = self.order
        node_dict["shape"] = self.shape
        return node_dict



# -*- coding: utf-8 -*-
"""
npxpy
Created on Thu Feb 29 11:49:17 2024

@author: Caghan Uenlueer
Neuromorphic Quantumphotonics
Heidelberg University
E-Mail:	caghan.uenlueer@kip.uni-heidelberg.de

This file is part of npxpy, which is licensed under the MIT License.
"""
import uuid
import copy
import os
from typing import Dict, Any, List, Tuple, Optional, Union
from importlib.resources import files


class Node:
    """
    A class to represent a node object of nanoPrintX with various attributes
    and methods for managing node hierarchy.

    Attributes:
        type (str): Type of the node.
        name (str): Name of the node.
        position (List[float]): Position of the node [x, y, z].
        rotation (List[float]): Rotation of the node [psi, theta, phi].
        children (List[str]): List of children node IDs.
        children_nodes (List[Node]): List of children nodes.
        properties (Any): Properties of the node.
        geometry (Any): Geometry of the node.
        unique_attributes (Dict[str, Any]): Additional dynamic attributes.
    """

    def __init__(
        self,
        node_type: str,
        name: str,
        position: List[float] = [0.0, 0.0, 0.0],
        rotation: List[float] = [0.0, 0.0, 0.0],
    ):
        """
        Initialize a Node instance with the specified parameters.

        Parameters:
            node_type (str): Type of the node.
            name (str): Name of the node.
            properties (Any, optional): Properties of the node. Defaults to None.
            geometry (Any, optional): Geometry of the node. Defaults to None.
            **kwargs (Any): Additional dynamic attributes.
        """

        self.id = str(uuid.uuid4())
        self._type = node_type
        self.name = name
        self.position = position
        self.rotation = rotation
        self.properties = {}
        self.geometry = {}

        self.children: List[str] = []
        self.children_nodes: List[Node] = []
        self.all_descendants: List[Node] = self._generate_all_descendants()

        self.parent_node: List[Node] = []
        self.all_ancestors: List[Node] = []

    @property
    def name(self):
        """Return the name of the node."""
        return self._name

    @property
    def node_type(self):
        """Return the type of the node."""
        return self._type

    @name.setter
    def name(self, value: str):
        value = str(value)
        """Set the name of the node with validation to ensure it is a non-empty string."""
        if not isinstance(value, str) or not value.strip():
            raise ValueError("name must be a non-empty string.")
        self._name = value

    @property
    def _visibility_in_plotter_disabled(self):
        return self.__visibility_in_plotter_disabled

    @_visibility_in_plotter_disabled.setter
    def _visibility_in_plotter_disabled(self, value):
        # Check if value is already an iterable (but not a string!)
        # If it's not an iterable (or is just a single string), wrap it in a list.
        if isinstance(value, str) or not hasattr(value, "__iter__"):
            value = [value]

        # Optionally convert it to a list if it's, say, a tuple or another iterable
        self.__visibility_in_plotter_disabled = list(value)

    def add_child(self, *child_nodes: "Node"):
        """
        Add child node(s) to the current node.

        Parameters:
            child_node (Node): The child node(s) to add.
        """
        for child_node in child_nodes:
            if not all(
                hasattr(child_node, attr)
                for attr in ["_type", "all_descendants", "all_ancestors"]
            ):
                raise TypeError(
                    "Only instances of nodes can be added as children to nodes!"
                )
            elif self._type == "structure":
                raise ValueError(
                    "Structure, Text and Lens are terminal nodes! They cannot have children!"
                )
            elif child_node._type == "project":
                raise ValueError(
                    "A project node can never be a child to any node!"
                )
            elif child_node._type == "scene":
                if self._has_ancestor_of_type("scene"):
                    raise ValueError("Nested scenes are not allowed!")
            elif self in child_node.all_descendants:
                raise ValueError(
                    "This node cannot be added since it is a ancestor to the current node!"
                )

            child_node.parent_node.append(self)
            self.children_nodes.append(child_node)
            # Update descendants list of parent

            self.all_descendants += [child_node] + child_node.all_descendants

            for child in [child_node] + child_node.all_descendants:
                child.all_ancestors = (
                    child._generate_all_ancestors()
                )  # Update ancestors list (_generate_all_ancestors() inexpensive and easy!)

            for (
                ancestor
            ) in (
                self.all_ancestors
            ):  # Update descendants for the parent's ancestors
                ancestor.all_descendants += [
                    child_node
                ] + child_node.all_descendants

            """
            # Deprecated?
            self.all_descendants = (
                self._generate_all_descendants()
            )  # Update descendants list
            child_node.all_ancestors = (
                child_node._generate_all_ancestors()
            )  # Update ancestors list
    
            for i in (
                self.all_descendants + child_node.all_ancestors
            ):  # Update for the whole batch of nodes their ancestors and descendants
                i.all_descendants = i._generate_all_descendants()
                i.all_ancestors = i._generate_all_ancestors()
            """
            if child_node._type == "structure":
                if (
                    not "scene" in [i._type for i in self.all_ancestors]
                    and "scene" != self._type
                ):
                    print("WARNING: Structures have to be inside Scene nodes!")

        return self

    def _has_ancestor_of_type(self, node_type: str) -> bool:
        """
        Check if the current node has an ancestor of the specified type.

        Parameters:
            node_type (str): The type of the ancestor node to check for.

        Returns:
            bool: True if an ancestor of the specified type exists, False otherwise.
        """
        current_node = self
        while current_node:
            if current_node._type == node_type:
                return True
            current_node = getattr(
                current_node, "parent", None
            )  # Assumes a parent attribute is set for each node
        return False

    def tree(
        self,
        level: int = 0,
        show_type: bool = True,
        show_id: bool = False,
        is_last: bool = True,
        prefix: str = "",
    ):
        """
        Print the tree structure of the node and its descendants.

        Parameters:
            level (int, optional): The current level in the tree. Defaults to 0.
            show_type (bool, optional): Whether to show the node type. Defaults to True.
            show_id (bool, optional): Whether to show the node ID. Defaults to False.
            is_last (bool, optional): Whether the node is the last child. Defaults to True.
            prefix (str, optional): The prefix for the current level. Defaults to ''.
        """
        indent = (
            "" if level == 0 else prefix + ("└" if is_last else "├") + "──"
        )
        output = (
            f"{indent}{self.name} ({self._type})"
            if show_type
            else f"{indent}{self.name}"
        )
        if show_id:
            output += f" (ID: {self.id})"
        print(output)
        new_prefix = prefix + ("    " if is_last else "│   ")
        child_count = len(self.children_nodes)
        for index, child in enumerate(self.children_nodes):
            child.tree(
                level + 1,
                show_type,
                show_id,
                is_last=(index == child_count - 1),
                prefix=new_prefix,
            )

    def deepcopy_node(self, copy_children: bool = True, name=None) -> "Node":
        """
        Create a deep copy of the node.

        Parameters:
            copy_children (bool, optional): Whether to copy children nodes. Defaults to True.

        Returns:
            Node: A deep copy of the current node.
        """

        copied_node = copy.copy(self)
        copied_node.id = str(uuid.uuid4())
        copied_node.children_nodes = []
        copied_node.all_descendants = []
        copied_node.parent_node = []
        copied_node.all_ancestors = []

        if copy_children:
            copied_children = [
                child.deepcopy_node() for child in self.children_nodes
            ]
            copied_node.add_child(*copied_children)

        if name is not None:
            copied_node.name = name
        return copied_node

    def _reset_ids(self, node: "Node"):
        """
        Reset the IDs of the node and its descendants.

        Parameters:
            node (Node): The node to reset IDs for.
        """
        node.id = str(uuid.uuid4())
        for child in node.children_nodes:
            self._reset_ids(child)

    def grab_node(self, *node_types_with_indices: Tuple[str, int]) -> "Node":
        """
        Grab nodes based on the specified types and indices.

        Parameters:
            node_types_with_indices (Tuple[str, int]):
            Tuples of arbitrary amount containing node type and index.

        Returns:
            Node: The node found based on the specified types and indices.
        """
        current_level_nodes = [self]
        for node_type, index in node_types_with_indices:
            next_level_nodes = []
            for node in current_level_nodes:
                filtered_nodes = [
                    child
                    for child in node.children_nodes
                    if child._type == node_type
                ]
                if len(filtered_nodes) > index:
                    next_level_nodes.append(filtered_nodes[index])
            current_level_nodes = next_level_nodes
        return current_level_nodes[0]

    def _generate_all_descendants(self) -> List["Node"]:
        """
        Generate a list of all descendant nodes.

        Returns:
            List[Node]: List of all descendant nodes.
        """
        descendants = []
        nodes_to_check = [self]
        while nodes_to_check:
            current_node = nodes_to_check.pop()
            descendants.extend(current_node.children_nodes)
            nodes_to_check.extend(current_node.children_nodes)
        return descendants

    def _generate_all_ancestors(self) -> List["Node"]:
        """
        Generate a list of all ancestor nodes.

        Returns:
            List[Node]: List of all descendant nodes.
        """
        ancestors = []
        nodes_to_check = [self]
        while nodes_to_check:
            current_node = nodes_to_check.pop()
            ancestors.extend(current_node.parent_node)
            nodes_to_check.extend(current_node.parent_node)
        return ancestors

    def grab_all_nodes_bfs(self, node_type: str) -> List["Node"]:
        """
        Grab all nodes of the specified type using breadth-first search.

        Parameters:
            node_type (str): The type of nodes to grab.

        Returns:
            List[Node]: List of nodes of the specified type.
        """
        result = []
        nodes_to_check = [self]
        while nodes_to_check:
            current_node = nodes_to_check.pop(0)  # Dequeue from the front
            if current_node._type == node_type:
                result.append(current_node)
            nodes_to_check.extend(
                current_node.children_nodes
            )  # Enqueue children
        return result

    def append_node(self, *nodes_to_append: "Node"):
        """
        Append a node to the deepest descendant on the highest branch.

        Parameters:
            node_to_append (Node): The node to append.
        """
        for node_to_append in nodes_to_append:
            grandest_grandchild = self._find_grandest_grandchild(self)
            grandest_grandchild.add_child(node_to_append)

    def _find_grandest_grandchild(self, current_node: "Node") -> "Node":
        """
        Find the deepest descendant node.

        Parameters:
            current_node (Node): The current node to start the search from.

        Returns:
            Node: The deepest descendant node.
        """
        if not current_node.children_nodes:
            return current_node
        else:
            grandest_children = [
                self._find_grandest_grandchild(child)
                for child in current_node.children_nodes
            ]
            return max(grandest_children, key=lambda node: self._depth(node))

    def _depth(self, node: "Node") -> int:
        """
        Calculate the depth of a node.

        Parameters:
            node (Node): The node to calculate the depth for.

        Returns:
            int: The depth of the node.
        """
        depth = 0
        current = node
        while current.children_nodes:
            current = current.children_nodes[0]
            depth += 1
        return depth

    def _lazy_import_wrapper(self):
        from . import _viewport_helpers

        return _viewport_helpers._lazy_import()

    def viewport(
        self,
        title: Optional[str] = None,
        disable: Optional[Union[str, List[str]]] = None,
        include_ancestor_transforms: Optional[bool] = True,
    ):
        """
        Opens a PyVista viewport visualizing the attached objects in this node.
        Notes
        -----
        - Does not visualize multiplications caused by Arrays yet.
        - Lenses might not be displayed accurately in the viewport visualization.
          However, this does not affect the final printed output.
        - Supports multiple node types such as `scene`, `structure`, `coarse_alignment`,
          `interface_alignment`, `fiber_core_alignment`, `marker_alignment`, `edge_alignment`,
          and `dose_compensation`.

        Parameters
        ----------
        title : str, optional
            Title to display in the PyVista window. Defaults to the name of the calling node.
        disable : str or list of str, optional
            One or more group names to disable visibility for. These groups will not be visible
            in the viewport. For example, `"scene"` or `["scene", "coarse_alignment"]`.
        include_ancestor_transforms : bool, optional
            Whether to apply transformations (position and rotation) from ancestor nodes to the
            visualized objects. Defaults to `True`.

        Returns
        -------
        _GroupedPlotter
            A customized plotter instance after rendering all meshes, including any
            transformations or visibility settings applied.

        Examples
        --------
        ```python
        >>> # Basic usage without disabling any groups
        >>> node.viewport()

        >>> # Disable visibility for "scene" and "coarse_alignment" groups
        >>> node.viewport(disable=["scene", "coarse_alignment"])

        >>> # Exclude transformations from ancestor nodes
        >>> node.viewport(include_ancestor_transforms=False)
        ```
        """
        _GroupedPlotter, _apply_transforms, _meshbuilder, blocks = (
            self._lazy_import_wrapper()
        )

        block_dicts_list = []

        if disable is None:
            disable = []
        if title is None:
            title = self.name
        # Create the plotter
        plotter = _GroupedPlotter(
            title=f"npxpy - Project Viewport ({title})",
            update_app_icon=False,
        )

        # Grid, axes, background
        plotter.show_grid(
            grid="back",
            location="outer",
            color="gray",
            xtitle="x",
            ytitle="y",
            ztitle="z",
            show_zlabels=True,
            padding=0.1,
            font_size=8,
        )
        plotter.show_axes()
        plotter.view_isometric()
        plotter.set_background("white")

        # Add logo widget
        logo_path = str(files("npxpy.images").joinpath("logo.png"))
        plotter.add_logo_widget(
            logo=logo_path,
            opacity=0.75,
            size=(0.15, 0.15),
            position=(0.84, 0.86),
        )
        # init the meshbuilder
        meshbuilder = _meshbuilder()

        for node in [self] + self.all_descendants:
            all_rotations = [
                getattr(ancestor, "rotation", [0, 0, 0])
                for ancestor in node.all_ancestors
            ]
            all_rotations.reverse()

            all_positions = [
                getattr(ancestor, "position", [0, 0, 0])
                for ancestor in node.all_ancestors
            ]
            all_positions.reverse()

            # If False, only go as far as node calling viewport()
            if not include_ancestor_transforms:
                if node != self:
                    dummy = node.all_ancestors.copy()
                    dummy.reverse()
                    start = dummy.index(self)
                    all_positions = all_positions[start:]
                    all_rotations = all_rotations[start:]
                else:
                    all_positions = [[0, 0, 0]]
                    all_rotations = [[0, 0, 0]]

            if node._type == "scene":
                scene = node
                # Create the circle representing the scene
                if len(scene.all_ancestors) != 0 and hasattr(
                    scene.all_ancestors[-1], "objective"
                ):
                    ronin_node = False
                    objective = scene.all_ancestors[-1].objective
                else:
                    print(
                        (
                            f"{self.name} not attached to any project node.\n"
                            "Objective is thus assumed to be x63."
                        )
                    )
                    ronin_node = True
                    objective = ""  # dummy. wont do anything in this case

                scene_mesh, scene_mesh_dict = _meshbuilder.scene_mesh(
                    objective, ronin_node
                )
                # apply first all_ancestors' rots

                _apply_transforms(
                    scene_mesh,
                    all_positions=all_positions + [scene.position],
                    all_rotations=all_rotations + [scene.rotation],
                )

                # Add to plotter as 'scene' group
                plotter.add_mesh(scene_mesh, **scene_mesh_dict)
                # blocks.append(scene_mesh)
                # block_dicts_list.append(scene_mesh_dict)
            # Structure
            if node._type == "structure" and node._mesh:
                structure = node
                loaded_mesh = _meshbuilder.load_mesh(structure.mesh.file_path)
                # apply initial mesh transformation
                _apply_transforms(
                    mesh=loaded_mesh,
                    all_positions=[structure.mesh.translation],
                    all_rotations=[structure.mesh.rotation],
                )

                # apply all_ancestors' rots and afterwards structure rotation
                _apply_transforms(
                    mesh=loaded_mesh,
                    all_rotations=all_rotations + [structure.rotation],
                    all_positions=all_positions + [structure.position],
                )

                # Add to plotter
                plotter.add_mesh(
                    loaded_mesh, color=structure.color, group=structure._type
                )
                # blocks.append(loaded_mesh)
                # block_dicts_list.append(
                #    {"color": structure.color, "group": structure._type}
                # )
            # Text (structure)
            if node._type == "structure" and hasattr(node, "font_size"):
                text_node = node
                text_mesh, text_mesh_dict = _meshbuilder.txt_mesh(text_node)
                # apply all_ancestors' rots and afterwards structure rotation
                _apply_transforms(
                    mesh=text_mesh,
                    all_rotations=all_rotations + [text_node.rotation],
                    all_positions=all_positions + [text_node.position],
                )

                plotter.add_mesh(text_mesh, **text_mesh_dict)
                # blocks.append(text_mesh)
                # block_dicts_list.append(text_mesh_dict)
            # Lens (structure)
            elif (
                node._type == "structure"
                and not node._mesh
                and not hasattr(node, "font_size")
            ):
                lens = node
                geometry = lens.to_dict()["geometry"]
                del (
                    geometry["type"],
                    geometry["nr_radial_segments"],
                    geometry["nr_phi_segments"],
                )
                lens_mesh = meshbuilder.lens_mesh(**geometry)

                _apply_transforms(
                    mesh=lens_mesh,
                    all_rotations=all_rotations + [lens.rotation],
                    all_positions=all_positions + [lens.position],
                )

                lens_mesh_dict = {
                    "color": lens.color,
                    "group": "structure_lens",
                }
                plotter.add_mesh(lens_mesh, **lens_mesh_dict)
                # blocks.append(lens_mesh)
                # block_dicts_list.append(lens_mesh_dict)
            # Coarse aligners
            if node._type == "coarse_alignment":
                coarse_aligner = node
                coarse_anchor_mesh, coarse_anchor_mesh_dict = (
                    _meshbuilder.ca_mesh(coarse_aligner)
                )
                # for coarse_anchor_mesh in coarse_anchor_meshes:
                # apply all_ancestors' rots and afterwards mesh rotation
                _apply_transforms(
                    mesh=coarse_anchor_mesh,
                    all_rotations=all_rotations,
                    all_positions=all_positions,
                )

                plotter.add_mesh(coarse_anchor_mesh, **coarse_anchor_mesh_dict)
                # blocks.append(coarse_anchor_mesh)
                # block_dicts_list.append(coarse_anchor_mesh_dict)
            # interface aligners
            if node._type == "interface_alignment":
                interface_aligner = node

                ia_mesh, ia_mesh_dict = meshbuilder.ia_mesh(
                    interface_aligner_node=interface_aligner
                )

                _apply_transforms(
                    mesh=ia_mesh,
                    all_rotations=all_rotations,
                    all_positions=all_positions,
                )

                plotter.add_mesh(ia_mesh, **ia_mesh_dict)
                # blocks.append(ia_mesh)
                # block_dicts_list.append(ia_mesh_dict)
            # fiber aligners
            if node._type == "fiber_core_alignment":
                fiber_aligner = node
                fa_mesh, fa_mesh_dict = meshbuilder.fa_mesh(fiber_aligner)

                _apply_transforms(
                    mesh=fa_mesh,
                    all_rotations=all_rotations,
                    all_positions=all_positions,
                )

                plotter.add_mesh(fa_mesh, **fa_mesh_dict)
                # blocks.append(fa_mesh)
                # block_dicts_list.append(fa_mesh_dict)
            # Marker aligners
            if node._type == "marker_alignment":
                marker_aligner = node
                ma_meshes, ma_label_meshes, ma_mesh_dict, ma_label_dict = (
                    _meshbuilder.ma_mesh(marker_aligner)
                )

                for ma_mesh, ma_label_mesh in zip(ma_meshes, ma_label_meshes):

                    # apply all_ancestors' rots
                    _apply_transforms(
                        mesh=ma_mesh,
                        all_rotations=all_rotations,
                        all_positions=all_positions,
                    )
                    # apply all_ancestors' rots
                    _apply_transforms(
                        mesh=ma_label_mesh,
                        all_rotations=all_rotations,
                        all_positions=all_positions,
                    )

                    plotter.add_mesh(ma_mesh, **ma_mesh_dict)
                    # blocks.append(ma_mesh)
                    # block_dicts_list.append(ma_mesh_dict)
                    plotter.add_mesh(ma_label_mesh, **ma_label_dict)
                    # blocks.append(ma_label_mesh)
                    # block_dicts_list.append(ma_label_dict)

            #  Edge aligners
            if node._type == "edge_alignment":
                edge_aligner = node
                edge_aligner_meshes, edge_aligner_meshes_dicts = (
                    _meshbuilder.ea_mesh(edge_aligner)
                )

                for mesh, mesh_dict in zip(
                    edge_aligner_meshes, edge_aligner_meshes_dicts
                ):

                    _apply_transforms(
                        mesh,
                        all_positions=[edge_aligner.edge_location + [0]],
                        all_rotations=[
                            [0, 0, edge_aligner.edge_orientation]
                        ],  # (in-plane rotation only)
                    )

                    # apply all_ancestors' rots
                    _apply_transforms(
                        mesh,
                        all_rotations=all_rotations,
                        all_positions=all_positions,
                    )

                    plotter.add_mesh(mesh, **mesh_dict)
                    # blocks.append(mesh)
                    # block_dicts_list.append(mesh_dict)
            # Dose compensation
            if node._type == "dose_compensation":
                dose_compensation = node

                for mesh, kwargs in _meshbuilder.dc_meshes(
                    dose_compensation.domain_size
                ):
                    _apply_transforms(
                        mesh=mesh,
                        all_rotations=all_rotations
                        + [[0, 0, dose_compensation.edge_orientation]],
                        all_positions=all_positions
                        + [dose_compensation.edge_location],
                    )

                    plotter.add_mesh(mesh, **kwargs)
                    # blocks.append(mesh)
                    # block_dicts_list.append(kwargs)
            # Capture
            if node._type == "capture":
                capture = node

                capt_mesh, capt_mesh_dict = _meshbuilder.capture_mesh(capture)

                _apply_transforms(
                    mesh=capt_mesh,
                    all_rotations=all_rotations,
                    all_positions=all_positions,
                )

                plotter.add_mesh(capt_mesh, **capt_mesh_dict)
                # blocks.append(capt_mesh)
                # block_dicts_list.append(capt_mesh_dict)

        actor, mapper = plotter.add_composite(blocks)
        for idx, block_dict in enumerate(block_dicts_list, start=1):
            for key, value in block_dict.items():
                setattr(mapper.block_attr[idx], key, value)

        # Disable visibility for certain groups if requested
        self._visibility_in_plotter_disabled = disable
        for grp in self._visibility_in_plotter_disabled:
            plotter.disable(grp)

        plotter._add_custom_axes()
        # Show the viewport
        plotter.show()

        return plotter

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the node and its attributes to a dictionary format.

        Returns:
            Dict[str, Any]: Dictionary representation of the node.
        """
        self.children = [i.id for i in self.children_nodes]
        node_dict = {
            "type": self._type,
            "id": self.id,
            "name": self.name,
            "position": self.position,
            "rotation": self.rotation,
            "children": self.children,
            "properties": self.properties,
            # "geometry": self.geometry,
        }

        return node_dict



# -*- coding: utf-8 -*-
"""
npxpy.gds
Created on Fri Feb 21 15:54:24 2025

@author: Caghan Uenlueer
Neuromorphic Quantumphotonics
Heidelberg University
E-Mail:	caghan.uenlueer@kip.uni-heidelberg.de

This file is part of npxpy, which is licensed under the MIT License.
"""
import os
import math
import sys
from typing import List, Dict, Tuple, Optional, Callable
from io import StringIO
from functools import wraps
from inspect import signature
from npxpy import (
    Scene,
    Project,
    Mesh,
    Image,
    Structure,
    Preset,
    InterfaceAligner,
    CoarseAligner,
    MarkerAligner,
    Group,
)

# Dependency management with improved error reporting
_MISSING_DEPS: List[str] = []
_HAS_PYA = False

try:
    import pya

    _HAS_PYA = True
except ImportError as e:
    raise ImportError(
        "Missing required dependency: 'pya' (Python for KLayout API).\n"
        "Install with: pip install npxpy[gds]\n"
        "or: pip install npxpy[all]"
    ) from e

try:
    import numpy as np

    _HAS_NUMPY = True
except ImportError:
    np = None
    _MISSING_DEPS.append("numpy")

try:
    from shapely.geometry import Polygon, MultiPolygon, box
    from shapely.affinity import translate, rotate
    from shapely.ops import unary_union

    _HAS_SHAPELY = True
except ImportError:
    Polygon = MultiPolygon = box = translate = rotate = unary_union = None
    _MISSING_DEPS.append("shapely")

try:
    import trimesh

    _HAS_TRIMESH = True
except ImportError:
    trimesh = None
    _MISSING_DEPS.append("trimesh")

try:
    import PIL

    _HAS_PIL = True
except ImportError:
    PIL = None
    _MISSING_DEPS.append("PIL")


def verbose_output(verbose_param: str = "_verbose") -> Callable:
    """Decorator to suppress print statements based on verbosity flag.

    Args:
        verbose_param: Name of the verbosity parameter in the decorated function

    Returns:
        Callable: Decorated function with output suppression
    """

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs) -> Callable:
            sig = signature(func)
            bound_args = sig.bind(*args, **kwargs)
            bound_args.apply_defaults()
            verbose: bool = bound_args.arguments.get(verbose_param, False)

            original_stdout = sys.stdout
            if not verbose:
                sys.stdout = StringIO()

            try:
                result = func(*args, **kwargs)
                if not verbose:
                    sys.stdout.getvalue()  # Consume captured output
                return result
            finally:
                sys.stdout = original_stdout

        return wrapper

    return decorator


class _write_field_scene(Scene):
    """Custom scene configuration for write field operations."""

    def __init__(self) -> None:
        """Initialize a default write field scene with interface aligner."""
        super().__init__(name="default_write_field")
        ia = InterfaceAligner(signal_type="auto", measure_tilt=True)
        ia.set_grid(count=[2, 2], size=[180.0, 180.0])
        self.add_child(ia)


class GDSParser:
    """Parser for GDSII layout files with dependency management and validation.

    Attributes:
        gds_file (str): Path to the loaded GDSII file
        layout (pya.Layout): Parsed GDSII layout object
        gds_name (str): Base name of the GDS file without extension
    """

    _REQUIRED_DEPS: List[str] = ["numpy", "shapely", "trimesh", "PIL"]

    def __init__(self, gds_file: str) -> None:
        """Initialize GDS parser with file validation and dependency checks.

        Args:
            gds_file: Path to GDSII file to load

        Raises:
            ImportError: If required dependencies are missing
            FileNotFoundError: If specified file doesn't exist
            ValueError: For invalid file types or parsing errors
        """
        self.gds_file = gds_file  # Validated through property setter
        self._layout = pya.Layout()
        self._layout.read(gds_file)  # Let pya exceptions bubble up
        self._plot_tiles_flag = False
        self._previous_image_safe_path_marker_aligned_printing = "0/0"
        self._check_dependencies()

    def _check_dependencies(self) -> None:
        """Verify required dependencies are installed.

        Raises:
            ImportError: With list of missing dependencies
        """
        missing = [dep for dep in self._REQUIRED_DEPS if dep in _MISSING_DEPS]
        if missing:
            raise ImportError(
                f"Missing required dependencies: {', '.join(missing)}\n"
                "Install either with: pip install npxpy[gds]\n"
                "or with: pip install npxpy[all]"
            )

    @property
    def gds_file(self) -> str:
        """Get path to loaded GDS file."""
        return self._gds_file

    @property
    def layout(self) -> pya.Layout:
        """Get parsed GDS layout object."""
        return self._layout

    @property
    def gds_name(self) -> str:
        """Get base name of GDS file without extension."""
        base = os.path.basename(self.gds_file)
        return os.path.splitext(base)[0]

    @gds_file.setter
    def gds_file(self, value: str) -> None:
        """Validate and set new GDS file path.

        Args:
            value: Path to new GDS file

        Raises:
            TypeError: For non-string input
            FileNotFoundError: If file doesn't exist
            ValueError: For non-GDS file extension
        """
        if not isinstance(value, str):
            raise TypeError(f"Expected string path, got {type(value)}")

        norm_path = os.path.normpath(value)
        if not os.path.isfile(norm_path):
            raise FileNotFoundError(f"GDS file not found: {norm_path}")

        if not value.lower().endswith(".gds"):
            raise ValueError("File must have .gds extension")

        self._gds_file = norm_path

    def _gather_polygons_in_child_cell(self, child_cell, layer_to_print):
        """
        Return a list of NumPy arrays containing polygon coordinates in the given
        cell.
        """
        polygons = []
        for shape in child_cell.shapes(layer_to_print):
            if shape.is_polygon() or shape.is_box():
                poly = shape.dpolygon
                # Convert the polygon's points to a NumPy array
                coords = np.array([(p.x, p.y) for p in poly.each_point_hull()])
                polygons.append(coords)
        return polygons

    def _polygons_to_shapely(self, polygons_np):
        """
        Convert a list of NumPy arrays (each shape (N,2))
        into a list of shapely Polygons.
        """
        shapely_polygons = []
        for arr in polygons_np:
            # Ensure closure if needed, or let Shapely handle it
            # Note: If your arrays are not closed, Shapely still interprets them as closed
            shapely_polygons.append(Polygon(arr))
        return shapely_polygons

    def _tile_polygon(self, ix, iy, tile_size, epsilon):
        """
        Return a shapely Polygon for the tile at (ix, iy),
        where each tile is 100×100, and (0,0) tile is centered around the origin:
           => x in [ix*100 - 50, ix*100 + 50]
           => y in [iy*100 - 50, iy*100 + 50]
        """
        xmin = ix * tile_size - tile_size / 2 - epsilon
        xmax = ix * tile_size + tile_size / 2 + epsilon
        ymin = iy * tile_size - tile_size / 2 - epsilon
        ymax = iy * tile_size + tile_size / 2 + epsilon
        return box(xmin, ymin, xmax, ymax)  # shapely box

    def _get_bounding_box(self, shapely_polygons):
        """
        Returns (min_x, min_y, max_x, max_y) that bounds all given shapely polygons.
        """
        minx = min(poly.bounds[0] for poly in shapely_polygons)
        miny = min(poly.bounds[1] for poly in shapely_polygons)
        maxx = max(poly.bounds[2] for poly in shapely_polygons)
        maxy = max(poly.bounds[3] for poly in shapely_polygons)
        return (minx, miny, maxx, maxy)

    def _tile_indices_for_bounding_box(
        self, minx, miny, maxx, maxy, tile_size
    ):
        """
        Given a bounding box and tile size (100 by default),
        yield (ix, iy) indices that cover all polygons.

        We define tiles so that the tile at (0,0) covers x in [-50, 50], y in [-50, 50].
        That means for a tile index (ix, iy), the tile covers:
            x in [ix*100 - 50, ix*100 + 50]
            y in [iy*100 - 50, iy*100 + 50].
        """
        # Figure out what range of indices we need in x and y directions
        # We shift coordinates so that "center" tile is from -50 to 50, etc.
        # Solve for ix such that ix*100 - 50 <= minx  =>  ix <= (minx + 50)/100.
        # But we need integer steps. We'll take floor for start, ceil for end.

        # X range
        ix_min = math.floor(
            (minx + tile_size / 2) / tile_size
        )  # leftmost tile index
        ix_max = math.ceil(
            (maxx + tile_size / 2) / tile_size
        )  # rightmost tile index

        # Y range
        iy_min = math.floor(
            (miny + tile_size / 2) / tile_size
        )  # bottom tile index
        iy_max = math.ceil(
            (maxy + tile_size / 2) / tile_size
        )  # top tile index

        for ix in range(ix_min, ix_max):
            for iy in range(iy_min, iy_max):
                yield ix, iy

    def _clip_polygons_to_tiles(self, shapely_polygons, tile_size, epsilon):
        """
        Main routine:
          1) Find bounding box of all polygons
          2) Figure out which tiles we need
          3) For each tile, intersect with each polygon
          4) Collect non-empty intersections in a result dictionary

        Returns a dict: {
           (ix, iy): [list of clipped Polygons / MultiPolygons within that tile]
        }
        """
        # 1) bounding box
        minx, miny, maxx, maxy = self._get_bounding_box(shapely_polygons)

        # 2) gather tiles
        tile_dict = {}  # (ix, iy) -> list of shapely geometries
        for ix, iy in self._tile_indices_for_bounding_box(
            minx, miny, maxx, maxy, tile_size
        ):
            tile_poly = self._tile_polygon(ix, iy, tile_size, epsilon)
            # 3) intersect with each polygon
            clipped_list = []
            for poly in shapely_polygons:
                intersection = poly.intersection(tile_poly)
                if not intersection.is_empty:
                    clipped_list.append(intersection)
            # Store if we got any intersection
            if clipped_list:
                tile_dict[(ix, iy)] = clipped_list
        return tile_dict

    def _tile_polygons(self, polygons_np, tile_size, epsilon):
        # 1) Convert to shapely Polygons
        shapely_polys = self._polygons_to_shapely(polygons_np)

        # 2) Clip polygons to tiles
        tile_dict = self._clip_polygons_to_tiles(
            shapely_polys, tile_size=tile_size, epsilon=epsilon
        )

        # Print how many tiles we actually used
        print(f"Number of tiles with content: {len(tile_dict)}")
        for tile_idx, clipped_geoms in tile_dict.items():
            print(
                f"Tile {tile_idx} has {len(clipped_geoms)} clipped polygon(s)"
            )

        # OPTIONAL: Visualize result
        if self._plot_tiles_flag:
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots(figsize=(8, 8))

            # Draw each tile that has content
            for (ix, iy), geoms in tile_dict.items():
                tile_poly = self._tile_polygon(
                    ix, iy, tile_size=tile_size, epsilon=epsilon
                )
                # Draw tile boundary
                x_tile, y_tile = tile_poly.exterior.xy
                ax.plot(x_tile, y_tile, "k--", alpha=0.3)

                # Draw clipped polygons
                for geom in geoms:
                    if geom.geom_type == "Polygon":
                        x, y = geom.exterior.xy
                        ax.fill(x, y, alpha=0.5)
                        for hole in geom.interiors:
                            xh, yh = hole.xy
                            ax.fill(xh, yh, color="white")
                    elif geom.geom_type == "MultiPolygon":
                        for part in geom.geoms:
                            x, y = part.exterior.xy
                            ax.fill(x, y, alpha=0.5)
                            for hole in part.interiors:
                                xh, yh = hole.xy
                                ax.fill(xh, yh, color="white")

            ax.set_aspect("equal", "box")
            ax.set_xlabel("X")
            ax.set_ylabel("Y")
            ax.set_title(
                f"Polygons clipped to {tile_size}um x {tile_size}um tiles"
            )
            plt.grid(True)
            plt.show()

        return tile_dict

    def _extrude_shapely_geometry(self, geometry, thickness):
        """
        Extrude a Shapely geometry (Polygon or MultiPolygon) by 'thickness'
        along the Z-axis and return a Trimesh mesh.
        """
        meshes = []

        # geometry can be Polygon or MultiPolygon
        if geometry.geom_type == "Polygon":
            # Directly extrude
            mesh = trimesh.creation.extrude_polygon(geometry, thickness)
            meshes.append(mesh)

        elif geometry.geom_type == "MultiPolygon":
            # Extrude each sub-polygon
            for poly in geometry.geoms:
                mesh = trimesh.creation.extrude_polygon(poly, thickness)
                meshes.append(mesh)

        # Combine everything into one mesh
        if len(meshes) == 1:
            return meshes[0]
        elif len(meshes) > 1:
            return trimesh.util.concatenate(meshes)
        else:
            # In case geometry was empty, return None or an empty Trimesh
            return None

    def _tile_polygons_2D_extrusion(
        self, extrusion, tile_dict, child_cell, target_layer, skip_if_exists
    ):

        output_folder = f"{self.gds_name}/{child_cell.name}{target_layer}"
        os.makedirs(output_folder, exist_ok=True)

        for (ix, iy), geoms in tile_dict.items():
            # Generate the tile filename and path
            tile_filename = f"tile_{ix}_{iy}.stl"
            tile_filepath = os.path.join(output_folder, tile_filename)

            # Check if the STL file already exists
            if os.path.exists(tile_filepath) and skip_if_exists:
                print(
                    f"Tile {(ix, iy)} already exists at {tile_filepath}, skipping."
                )
                continue

            # List to collect meshes from each geometry
            tile_meshes = []

            # Extrude each geometry in that tile
            for geom in geoms:
                mesh_3d = self._extrude_shapely_geometry(
                    geometry=geom, thickness=extrusion
                )
                if mesh_3d is not None:
                    tile_meshes.append(mesh_3d)

            # Combine (concatenate) all extruded meshes in this tile
            if len(tile_meshes) == 0:
                # No valid geometry in this tile, skip
                continue
            elif len(tile_meshes) == 1:
                tile_mesh_combined = tile_meshes[0]
            else:
                tile_mesh_combined = trimesh.util.concatenate(tile_meshes)

            # Export to STL
            tile_mesh_combined.export(tile_filepath)
            print(f"Exported tile {(ix, iy)} to {tile_filepath}")

    def _meander_order(self, tile_keys):
        """
        Given an iterable of (ix, iy) tile indices,
        return a list of (ix, iy) in a zigzag (meander) order.

        - Sort by ascending y for the rows.
        - For each consecutive row, alternate the x-direction:
            * row0: left-to-right
            * row1: right-to-left
            * row2: left-to-right
            * ...
        """
        # Group the tile indices by their y
        from collections import defaultdict

        rows = defaultdict(list)
        for ix, iy in tile_keys:
            rows[iy].append(ix)

        # Sort the rows by Y ascending
        sorted_ys = sorted(rows.keys())

        # Build the final list of (ix, iy)
        meandered = []
        for row_i, y in enumerate(sorted_ys):
            x_list = sorted(rows[y])
            # If row_i is odd, reverse the list to create the zigzag
            if row_i % 2 == 1:
                x_list.reverse()

            for x in x_list:
                meandered.append((x, y))

        return meandered

    def _tile_center(self, ix, iy, tile_size):
        """
        Return the center of tile (ix, iy) in the same coordinate system
        that was used for clipping (i.e., each tile is tile_size wide/high).
        """
        cx = ix * tile_size
        cy = iy * tile_size
        return (cx, cy)

    def _build_nano_leaf_group(
        self,
        tile_dict,
        tile_size,
        project,
        preset,
        leaf_cell,
        layer_to_print,
        group_xy,
        rotation,
        write_field_scene=None,
        color="#16506B",
    ):
        # 1) Collect tile keys and meander them
        tile_keys = list(tile_dict.keys())
        meandered_keys = self._meander_order(tile_keys)
        if write_field_scene is None:
            write_field_scene = _write_field_scene()
        else:  # Sleep-deprived much? Time to go to bed...
            try:
                if write_field_scene._type != "scene":
                    raise TypeError(
                        "write_field_scene needs to be of node type scene"
                    )
            except:
                write_field_scene = _write_field_scene()
                UserWarning(
                    "Invalid scene. Default write field going to be applied instead."
                )
        # 2) Build Scenes in meander order
        scenes = []
        meshes = []
        for i, (ix, iy) in enumerate(meandered_keys):
            # tile_{ix}_{iy}.stl is our new naming scheme
            stl_filename = f"{self.gds_name}/{leaf_cell.name}{layer_to_print}/tile_{ix}_{iy}.stl"

            # Compute the tile center
            cx, cy = self._tile_center(ix, iy, tile_size=tile_size)

            # Build the Scene, position = [cx, cy, 0]
            scene = write_field_scene.deepcopy_node(
                name=stl_filename
            ).position_at(position=[cx, cy, 0])

            # Build the Mesh object
            # auto_center=True => internally centers the geometry
            mesh_obj = Mesh(
                stl_filename, name=stl_filename, translation=[-cx, -cy, 0]
            )

            # Prepare for the structure
            # (assuming you have your 'preset' object already loaded)
            structure = Structure(
                name=mesh_obj.name, preset=preset, mesh=mesh_obj, color=color
            )

            # Attach structure to the scene
            scene.append_node(structure)

            # Keep references
            scenes.append(scene)
            meshes.append(mesh_obj)

        project.load_resources(meshes)
        leaf_group = Group(
            name=leaf_cell.name,
            position=[*group_xy, 0],
            rotation=[0, 0, rotation],
        )
        leaf_group.add_child(*scenes)
        return leaf_group

    def _cell_has_direct_polygons(
        self, cell: pya.Cell, layer_to_print: int
    ) -> bool:
        """
        Check if a cell directly contains any polygon shapes on a specific layer.

        Args:
            cell: pya.Cell to check
            layer_to_print: Layer index to examine

        Returns:
            True if cell directly contains polygons on this layer, False otherwise
        """
        for shape in cell.shapes(layer_to_print):
            if shape.is_polygon() or shape.is_box():
                return True
        return False

    groups = []

    def _collect_instance_displacements(self, cell):
        displacements = []
        dbu = self.layout.dbu  # Database unit to micron conversion factor

        for instance in cell.each_inst():
            # Extract array parameters (default to 1 if not an array)
            na = instance.na or 1
            nb = instance.nb or 1
            a_vec = (
                instance.a
            )  # Column displacement vector (in database units)
            b_vec = instance.b  # Row displacement vector

            # Base displacement from the instance's transformation
            base_disp_db = instance.trans.disp  # In database units

            # Iterate over all elements in the array
            for i in range(na):
                for j in range(nb):
                    # Compute total displacement for this array element
                    delta = a_vec * i + b_vec * j
                    total_disp_db = base_disp_db + delta

                    # Convert to microns and add to the list
                    total_disp_micron = total_disp_db.to_dtype(dbu)
                    displacements.append(
                        [total_disp_micron.x, total_disp_micron.y]
                    )

        return displacements

    def gds_printing(
        self,
        project: Project,
        preset: Preset,
        cell_name: Optional[str] = None,
        write_field_scene: Optional[Scene] = None,
        layer_to_print: Tuple[int, int] = (1, 1),
        extrusion: float = 1.0,
        tile_size: float = 220,
        epsilon: float = 0.5,
        skip_if_exists: bool = False,
        color: str = "#16506B",
        _verbose: bool = False,
    ) -> Group:
        """Generate hierarchical printing structure from GDS layout polygons.

        Args:
            project: Target project for resource management
            preset: Printing preset configuration
            cell_name: Specific cell to process (uses top cell if None)
            write_field_scene: Custom write field scene configuration
            layer_to_print: Layer/datatype tuple to process
            extrusion: Z-height for printed structures (negative values valid!)
            tile_size: Maximum dimension for geometry tiling
            epsilon: Overlap-compensation for stitching in microns
            skip_if_exists: If polygons already exist, do not recreate them
            color: Visualization color of meshes inside viewport
            _verbose: Enable debug output

        Returns:
            Group: Hierarchical structure ready for printing

        Raises:
            ValueError: Invalid input parameters
            TypeError: Incorrect argument types
            RuntimeError: Polygon processing failure
        """
        # Input validation
        if not isinstance(project, Project):
            raise TypeError("project must be a Project instance")
        if not isinstance(preset, Preset):
            raise TypeError("preset must be a Preset instance")
        if cell_name is not None and not isinstance(cell_name, str):
            raise TypeError("cell_name must be a string or None")
        if write_field_scene is not None and not isinstance(
            write_field_scene, Scene
        ):
            raise TypeError(
                "write_field_scene must be a Scene instance or None"
            )

        # Validate layer_to_print structure and content
        if not isinstance(layer_to_print, tuple) or len(layer_to_print) != 2:
            raise TypeError("layer_to_print must be a tuple of two integers")
        if not all(isinstance(x, int) for x in layer_to_print):
            raise TypeError("Both elements in layer_to_print must be integers")

        # Validate numerical parameters
        if not isinstance(extrusion, (int, float)):
            raise TypeError("extrusion must be a numeric value")
        if not isinstance(tile_size, (int, float)):
            raise TypeError("tile_size must be a numeric value")
        if tile_size <= 0:
            raise ValueError("tile_size must be positive")
        if not isinstance(epsilon, (int, float)):
            raise TypeError("epsilon must be a numeric value")
        if epsilon < 0:
            raise ValueError("epsilon must be non-negative")

        # Validate boolean parameters
        if not isinstance(skip_if_exists, bool):
            raise TypeError("skip_if_exists must be a boolean")
        if not isinstance(_verbose, bool):
            raise TypeError("_verbose must be a boolean")

        gds_printing_group_raw = self._gds_printing(
            project,
            preset,
            cell_name=cell_name,
            write_field_scene=write_field_scene,
            layer_to_print=layer_to_print,
            extrusion=extrusion,
            tile_size=tile_size,
            epsilon=epsilon,
            skip_if_exists=skip_if_exists,
            color=color,
            _verbose=_verbose,
        )

        # Clean up nodes that do not contain any structures
        gds_printing_group = gds_printing_group_raw.deepcopy_node(
            copy_children=False
        )
        for node in gds_printing_group_raw.children_nodes:
            for node_descendant in node.all_descendants:
                if node_descendant._type == "structure":
                    gds_printing_group.add_child(node)
                    break
        return gds_printing_group

        return gds_printing_group

    @verbose_output()
    def _gds_printing(
        self,
        project: Project,
        preset: Preset,
        cell_name: Optional[str],
        write_field_scene: Optional[Scene],
        layer_to_print: Tuple[int, int],
        extrusion: float,
        tile_size: float,
        epsilon: float,
        skip_if_exists: bool,
        color: str,
        _verbose: bool,
    ) -> Group:
        cell = (
            self.layout.top_cell()
            if cell_name is None
            else self.get_cell_by_name(cell_name)
        )
        print(f"Cell: {cell.name}")
        cell_group = Group(f"Cell: {cell.name} Layer:{layer_to_print}")
        for instance in cell.each_inst():

            # Get the child cell
            child_cell = self.layout.cell(instance.cell_index)

            # Get the transformation of the instance
            trans = instance.trans

            # Extract the displacement vector (relative translation)
            displacement = trans.disp
            rotation = (
                trans.rot * 90
            )  # outputs are ints (0,1,2,3) for multiples of 90 deg
            # Convert the displacement to microns (if needed)
            displacement_in_microns = displacement.to_dtype(self.layout.dbu)

            print(f"Child cell: {child_cell.name}")
            # print(f"Relative displacement (in database units): {displacement}")
            print(
                f"Relative displacement (in microns): {displacement_in_microns.x, displacement_in_microns.y}"
            )
            print("---")

            if self._cell_has_direct_polygons(child_cell, layer_to_print):
                polygons = self._gather_polygons_in_child_cell(
                    child_cell, layer_to_print
                )
                tile_dict = self._tile_polygons(
                    polygons, tile_size=tile_size, epsilon=epsilon
                )
                self._tile_polygons_2D_extrusion(
                    extrusion=extrusion,
                    tile_dict=tile_dict,
                    child_cell=child_cell,
                    target_layer=layer_to_print,
                    skip_if_exists=skip_if_exists,
                )
                child_cell_group = self._build_nano_leaf_group(
                    tile_dict,
                    tile_size,
                    project,
                    preset,
                    child_cell,
                    group_xy=[
                        displacement_in_microns.x,
                        displacement_in_microns.y,
                    ],
                    rotation=rotation,
                    write_field_scene=write_field_scene,
                    layer_to_print=layer_to_print,
                    color=color,
                )

            else:
                child_cell_group = Group(
                    name=child_cell.name,
                    position=[
                        displacement_in_microns.x,
                        displacement_in_microns.y,
                        0,
                    ],
                    rotation=[0, 0, rotation],
                )
                print("No direct polygons found in top cell")

            #  Do NOT assume you could shove this in the if-statement above
            if not child_cell.is_leaf():
                child_cell_group.add_child(
                    self.gds_printing(
                        project,
                        preset,
                        cell=child_cell,
                        write_field_scene=write_field_scene,
                        layer_to_print=layer_to_print,
                        extrusion=extrusion,
                        tile_size=tile_size,
                        epsilon=epsilon,
                        color=color,
                        _verbose=_verbose,
                    )
                )
            else:
                cell_group.add_child(child_cell_group)

                print("LEAF!")

        return cell_group

    def _decompose(self, geometry):
        """Decompose a geometry into a list of Polygon(s)."""
        if isinstance(geometry, MultiPolygon):
            return list(geometry.geoms)
        elif isinstance(geometry, Polygon):
            return [geometry]
        else:
            raise ValueError("Unsupported geometry type")

    def _get_polygon_coords(self, polygon):
        """Extract all coordinates from a polygon (exterior and interiors)."""
        exterior = list(polygon.exterior.coords)
        interiors = []
        for interior in polygon.interiors:
            interiors.extend(interior.coords)
        return np.array(exterior + interiors)

    def _normalize_polygon(self, polygon):
        """Normalize a polygon's position, rotation, and orientation."""
        # Translate to centroid origin
        centroid = polygon.centroid
        translated = translate(polygon, -centroid.x, -centroid.y)

        # Get coordinates for PCA
        coords = self._get_polygon_coords(translated)
        if len(coords) < 2:
            return translated  # Not enough points for PCA

        # Compute PCA to find the principal axis
        centered = coords - np.mean(coords, axis=0)
        cov = np.cov(centered.T)
        eigenvalues, eigenvectors = np.linalg.eig(cov)
        principal = eigenvectors[:, np.argmax(eigenvalues)]
        angle = np.arctan2(principal[1], principal[0])

        # Rotate to align principal axis with x-axis
        rotated = rotate(translated, -np.degrees(angle), origin=(0, 0))

        # Heuristic to ensure consistent orientation (flip if necessary)
        coords_rotated = list(rotated.exterior.coords)
        if len(coords_rotated) >= 2:
            dx = coords_rotated[1][0] - coords_rotated[0][0]
            dy = coords_rotated[1][1] - coords_rotated[0][1]
            if dx < 0 or (dx == 0 and dy < 0):
                # Reflect across x-axis
                return rotate(rotated, 180, origin=(0, 0))
        return rotated

    def _are_geometries_equivalent(self, geom1, geom2, tolerance=1e-6):
        """Check if two geometries are equivalent in shape and size."""
        # Decompose into individual polygons
        polys1 = self._decompose(geom1)
        polys2 = self._decompose(geom2)
        if len(polys1) != len(polys2):
            return False

        # Normalize and sort polygons for comparison
        def sort_key(p):
            return (-p.area, -p.length, list(p.exterior.coords))

        normalized1 = sorted(
            [self._normalize_polygon(p) for p in polys1], key=sort_key
        )
        normalized2 = sorted(
            [self._normalize_polygon(p) for p in polys2], key=sort_key
        )

        # Compare each pair of polygons
        for p1, p2 in zip(normalized1, normalized2):
            if not p1.equals_exact(p2, tolerance):
                return False
        return True

    def _merge_touching_polygons(self, polygons):
        """
        Merge polygons that touch or intersect, including newly formed ones.
        Returns a list of merged geometries (Polygon/MultiPolygon).
        """
        processed = [False] * len(polygons)
        result = []

        for i in range(len(polygons)):
            if not processed[i]:
                # Start a new connected component
                component = [polygons[i]]
                processed[i] = True
                queue = [i]

                # Find all connected polygons using BFS
                while queue:
                    current_idx = queue.pop(0)
                    current_poly = polygons[current_idx]

                    # Check against all other polygons
                    for j in range(len(polygons)):
                        if not processed[j]:
                            other_poly = polygons[j]
                            if current_poly.intersects(other_poly):
                                component.append(other_poly)
                                processed[j] = True
                                queue.append(j)

                # Merge the component into a single geometry
                merged = unary_union(component)
                result.append(merged)

        return result

    def _ensure_folder_exist_else_create(self, path):
        try:
            if os.path.exists(path):
                pass
            else:
                os.makedirs(path)
        except Exception as e:
            print(f"An error occurred: {e}")

    def marker_aligned_printing(
        self,
        project: Project,
        presets: List[Preset],
        meshes: List[Mesh],
        marker_height: float = 0.33,
        marker_layer: Tuple[int, int] = (10, 10),
        mesh_spots_layers: List[Tuple[int, int]] = [(100, 100)],
        cell_origin_offset: Tuple[float, float] = (0.0, 0.0),
        cell_name: Optional[str] = None,
        image_resource: Optional[Image] = None,
        interface_aligner_node: Optional[InterfaceAligner] = None,
        marker_aligner_node: Optional[MarkerAligner] = None,
        colors: Optional[List[str]] = None,
        marker_aligner_kwargs: Optional[Dict] = None,
        structure_kwargs: Optional[Dict] = None,
        _verbose: bool = False,
    ) -> Group:
        """Create a hierarchical printing group with marker-based alignment.

        Args:
            project: Parent Project for resource management
            presets: List of Preset configurations for printing
            meshes: List of Mesh objects to print
            marker_height: Z-height for marker structures
            marker_layer: Layer/datatype for alignment markers
            mesh_spots_layers: List of layers containing print locations
            cell_origin_offset: Coordinate offset for cell origin
            cell_name: Cell to start traversing from (uses top cell if None)
            image_resource: Pre-configured Image resource for markers
            interface_aligner_node: InterfaceAligner configuration template
            marker_aligner_node: MarkerAligner configuration template
            colors: Color codes for visualization
            marker_aligner_kwargs: Additional MarkerAligner parameters
            structure_kwargs: Additional Structure parameters
            _verbose: Enable debug output

        Returns:
            Group: Hierarchical printing structure with alignment

        Raises:
            ValueError: Invalid input dimensions, values, or formats
            TypeError: Incorrect argument types
            RuntimeError: Marker processing failure
        """
        # Initialize mutable defaults safely
        marker_aligner_kwargs = marker_aligner_kwargs or {}
        structure_kwargs = structure_kwargs or {}
        colors = colors or ["#16506B"] * len(meshes)

        # Comprehensive type validation
        if not isinstance(project, Project):
            raise TypeError("project must be a Project instance")
        if not isinstance(presets, list):
            raise TypeError("presets must be a list")
        if not isinstance(meshes, list):
            raise TypeError("meshes must be a list")
        if not isinstance(mesh_spots_layers, list):
            raise TypeError("mesh_spots_layers must be a list")

        # Validate numerical parameters
        if not isinstance(marker_height, (int, float)):
            raise TypeError("marker_height must be numeric")
        if (
            not isinstance(cell_origin_offset, tuple)
            or len(cell_origin_offset) != 2
        ):
            raise TypeError("cell_origin_offset must be a 2-element tuple")
        if not all(isinstance(x, (int, float)) for x in cell_origin_offset):
            raise TypeError("cell_origin_offset elements must be numeric")

        # Validate layer specifications
        layer_valid = (
            lambda l: isinstance(l, tuple)
            and len(l) == 2
            and all(isinstance(n, int) for n in l)
        )
        if not layer_valid(marker_layer):
            raise TypeError("marker_layer must be a (int, int) tuple")
        if not all(layer_valid(l) for l in mesh_spots_layers):
            raise TypeError(
                "All mesh_spots_layers elements must be (int, int) tuples"
            )

        # Validate list contents
        for i, preset in enumerate(presets):
            if not isinstance(preset, Preset):
                raise TypeError(f"presets[{i}] must be a Preset instance")
        for i, mesh in enumerate(meshes):
            if not isinstance(mesh, Mesh):
                raise TypeError(f"meshes[{i}] must be a Mesh instance")

        # Validate optional parameters
        if cell_name is not None and not isinstance(cell_name, str):
            raise TypeError("cell_name must be a string or None")
        if image_resource is not None and not isinstance(
            image_resource, Image
        ):
            raise TypeError("image_resource must be an Image instance or None")
        if interface_aligner_node is not None and not isinstance(
            interface_aligner_node, InterfaceAligner
        ):
            raise TypeError(
                "interface_aligner_node must be an InterfaceAligner instance or None"
            )
        if marker_aligner_node is not None and not isinstance(
            marker_aligner_node, MarkerAligner
        ):
            raise TypeError(
                "marker_aligner_node must be a MarkerAligner instance or None"
            )

        # Validate dictionary parameters
        if not isinstance(marker_aligner_kwargs, dict):
            raise TypeError("marker_aligner_kwargs must be a dictionary")
        if not isinstance(structure_kwargs, dict):
            raise TypeError("structure_kwargs must be a dictionary")
        if not isinstance(_verbose, bool):
            raise TypeError("_verbose must be a boolean")

        # Validate dimensional consistency
        if (
            len(presets) != len(meshes)
            or len(presets) != len(mesh_spots_layers)
            or len(presets) != len(colors)
        ):
            raise ValueError("All input lists must have equal length")
        if not presets:
            raise ValueError("At least one preset must be provided")

        try:
            marker_aligned_printing_group_raw = self._marker_aligned_printing(
                project,
                presets,
                meshes,
                cell_name=cell_name,
                cell_origin_offset=cell_origin_offset,
                image_resource=image_resource,
                interface_aligner_node=interface_aligner_node,
                marker_aligner_node=marker_aligner_node,
                marker_height=marker_height,
                marker_layer=marker_layer,
                mesh_spots_layers=mesh_spots_layers,
                colors=colors,
                marker_aligner_kwargs=marker_aligner_kwargs,
                structure_kwargs=structure_kwargs,
                _verbose=_verbose,
            )
        except Exception as e:
            raise RuntimeError("Marker alignment processing failed") from e
        # Clean up nodes that do not contain any structures
        marker_aligned_printing_group = (
            marker_aligned_printing_group_raw.deepcopy_node(
                copy_children=False
            )
        )
        for node in marker_aligned_printing_group_raw.children_nodes:
            for node_descendant in node.all_descendants:
                if node_descendant._type == "structure":
                    marker_aligned_printing_group.add_child(node)
                    break
        return marker_aligned_printing_group.translate(
            [-cell_origin_offset[0], -cell_origin_offset[1], 0]
        )

    # TODO: Consider exchanging marker part with get_marker_aligner()
    @verbose_output()
    def _marker_aligned_printing(
        self,
        project: Project,
        presets: List[Preset],
        meshes: List[Mesh],
        marker_height: float,
        marker_layer: Tuple[int, int],
        mesh_spots_layers: List[Tuple[int, int]],
        cell_origin_offset: Tuple[float, float],
        cell_name: Optional[str],
        image_resource: Optional[Image],
        interface_aligner_node: Optional[InterfaceAligner],
        marker_aligner_node: Optional[MarkerAligner],
        colors: List[str],
        marker_aligner_kwargs: Dict,
        structure_kwargs: Dict,
        _verbose: bool,
    ) -> Group:
        """Internal implementation of marker-aligned printing."""
        cell = (
            self.layout.top_cell()
            if cell_name is None
            else self.get_cell_by_name(cell_name)
        )
        print(f"Cell: {cell.name}")
        cell_group = Group(f"Cell: {cell.name} markers:{marker_layer}")
        for instance in cell.each_inst():

            # Get the child cell
            child_cell = self.layout.cell(instance.cell_index)

            # Get the transformation of the instance
            trans = instance.trans

            # Extract the displacement vector (relative translation)
            displacement = trans.disp
            rotation = (
                trans.rot * 90
            )  # outputs are ints (0,1,2,3) for multiples of 90 deg
            # Convert the displacement to microns (if needed)
            displacement_in_microns = displacement.to_dtype(self.layout.dbu)

            print(f"Child cell: {child_cell.name}")
            # print(f"Relative displacement (in database units): {displacement}")
            print(
                f"Relative displacement (in microns): {displacement_in_microns.x, displacement_in_microns.y}"
            )
            print(f"Rotation: {rotation}, type: {type(rotation)}")
            print("---")

            if self._cell_has_direct_polygons(child_cell, marker_layer):
                child_cell_group = Group(
                    name=child_cell.name,
                    position=[
                        displacement_in_microns.x,
                        displacement_in_microns.y,
                        0,
                    ],
                    rotation=[0, 0, rotation],
                )
                image_file_path = f"./images_{self.gds_name}_{marker_layer}/marker_{marker_layer}.png"
                self._ensure_folder_exist_else_create(
                    f"./images_{self.gds_name}_{marker_layer}"
                )
                scene = Scene(name=child_cell.name)

                polygons = self._gather_polygons_in_child_cell(
                    child_cell, marker_layer
                )
                shapely_polygons = self._polygons_to_shapely(polygons)
                marker_polygons = self._merge_touching_polygons(
                    shapely_polygons
                )
                _, marker_orientations = (
                    self._group_equivalent_polygons_and_output_image(
                        marker_polygons, file_path=image_file_path
                    )
                )

                _image = (
                    Image(name=f"{marker_layer}", file_path=image_file_path)
                    if image_resource is None
                    else image_resource
                )
                if (
                    self._previous_image_safe_path_marker_aligned_printing.split(
                        "/"
                    )[
                        1
                    ]
                    != _image.safe_path.split("/")[1]
                ):

                    self._image = (
                        Image(
                            name=f"{marker_layer}", file_path=image_file_path
                        )
                        if image_resource is None
                        else image_resource
                    )
                    self._previous_image_safe_path_marker_aligned_printing = (
                        self._image.safe_path
                    )
                    project.load_resources(self._image)

                marker_size = [
                    marker_polygons[0].bounds[2]
                    - marker_polygons[0].bounds[0],
                    marker_polygons[0].bounds[3]
                    - marker_polygons[0].bounds[1],
                ]
                marker_positions = [
                    [m_pol.centroid.x, m_pol.centroid.y, marker_height]
                    for m_pol in marker_polygons
                ]

                if "max_outliers" not in marker_aligner_kwargs:
                    marker_aligner_kwargs["max_outliers"] = (
                        len(marker_positions) - 3
                        if len(marker_positions) >= 3
                        else 0
                    )
                marker_aligner = (
                    MarkerAligner(
                        name=f"{marker_layer}",
                        image=self._image,
                        marker_size=marker_size,
                        **marker_aligner_kwargs,
                    )
                    if marker_aligner_node is None
                    else marker_aligner_node.deepcopy_node()
                )

                marker_aligner.set_markers_at(
                    positions=marker_positions,
                    orientations=marker_orientations,
                )

                for mesh, preset, mesh_spots_layer, color in zip(
                    meshes, presets, mesh_spots_layers, colors
                ):
                    if self._cell_has_direct_polygons(
                        child_cell, mesh_spots_layer
                    ):
                        mesh_spots_polygons = (
                            self._gather_polygons_in_child_cell(
                                child_cell, mesh_spots_layer
                            )
                        )
                        mesh_spots_shapely_polygons = (
                            self._polygons_to_shapely(mesh_spots_polygons)
                        )
                        structures = [
                            Structure(
                                mesh=mesh,
                                preset=preset,
                                name=mesh.name,
                                position=[
                                    mesh_spot_shapely_polygon.centroid.x,
                                    mesh_spot_shapely_polygon.centroid.y,
                                    0,
                                ],
                                color=color,
                                **structure_kwargs,
                            )
                            for mesh_spot_shapely_polygon in mesh_spots_shapely_polygons
                        ]
                        marker_aligner.add_child(*structures)

                interface_aligner = (
                    InterfaceAligner()
                    if interface_aligner_node is None
                    else interface_aligner_node.deepcopy_node()
                )

                cell_origin_offset_group = Group(
                    name="cell_origin_offset",
                    position=[
                        cell_origin_offset[0],
                        cell_origin_offset[1],
                        0,
                    ],
                )
                child_cell_group.append_node(
                    scene,
                    interface_aligner,
                    cell_origin_offset_group,
                    marker_aligner,
                )

            else:
                child_cell_group = Group(
                    name=child_cell.name,
                    position=[
                        displacement_in_microns.x,
                        displacement_in_microns.y,
                        0,
                    ],
                    rotation=[0, 0, rotation],
                )
                print("No direct polygons found in top cell")

            #  Do NOT assume you could shove this in the if-statement above
            if not child_cell.is_leaf():
                cell_group.add_child(child_cell_group)
                child_cell_group.add_child(
                    self.marker_aligned_printing(
                        project,
                        presets,
                        meshes,
                        cell=child_cell,
                        image_resource=image_resource,
                        interface_aligner_node=interface_aligner_node,
                        marker_aligner_node=marker_aligner_node,
                        marker_height=marker_height,
                        marker_layer=marker_layer,
                        mesh_spots_layers=mesh_spots_layers,
                        colors=colors,
                        marker_aligner_kwargs=marker_aligner_kwargs,
                        structure_kwargs=structure_kwargs,
                        _verbose=_verbose,
                    )
                )

            else:
                cell_group.add_child(child_cell_group)

                print("LEAF!")

        return cell_group

    def _get_geometry_coords(self, geometry):
        """Extract all coordinates from a geometry (Polygon or MultiPolygon)."""
        coords = []
        if isinstance(geometry, MultiPolygon):
            for polygon in geometry.geoms:
                exterior = list(polygon.exterior.coords)
                coords.extend(exterior)
                for interior in polygon.interiors:
                    coords.extend(interior.coords)
        elif isinstance(geometry, Polygon):
            exterior = list(geometry.exterior.coords)
            coords.extend(exterior)
            for interior in geometry.interiors:
                coords.extend(interior.coords)
        else:
            raise ValueError("Unsupported geometry type")
        return np.array(coords)

    def _normalize_geometry_with_rotation(self, geometry):
        """Normalize a geometry and return the normalized version and rotation applied."""
        centroid = geometry.centroid
        translated = translate(geometry, -centroid.x, -centroid.y)

        coords = self._get_geometry_coords(translated)
        if len(coords) < 2:
            return translated, 0.0

        centered = coords - np.mean(coords, axis=0)
        cov = np.cov(centered.T)
        eigenvalues, eigenvectors = np.linalg.eig(cov)
        principal = eigenvectors[:, np.argmax(eigenvalues)]
        angle_rad = np.arctan2(principal[1], principal[0])
        angle_deg = np.degrees(angle_rad)
        rotated1 = rotate(translated, -angle_deg, origin=(0, 0))

        # Check orientation
        if isinstance(rotated1, MultiPolygon):
            first_poly = rotated1.geoms[0]
            coords_rotated = list(first_poly.exterior.coords)
        else:
            coords_rotated = list(rotated1.exterior.coords)

        flip = False
        if len(coords_rotated) >= 2:
            dx = coords_rotated[1][0] - coords_rotated[0][0]
            dy = coords_rotated[1][1] - coords_rotated[0][1]
            if dx < 0 or (dx == 0 and dy < 0):
                flip = True

        if flip:
            rotated_final = rotate(rotated1, 180, origin=(0, 0))
            total_rotation = -angle_deg + 180
        else:
            rotated_final = rotated1
            total_rotation = -angle_deg

        return rotated_final, total_rotation

    def _group_equivalent_polygons_and_output_image(
        self, polygons, tolerance=1e-6, file_path="./images/marker.png"
    ):
        """
        Groups polygons into equivalence classes based on shape and size, ignoring position and rotation.
        Returns unique representatives and their relative orientations.
        """
        groups = []  # Each entry is (original_geo, rotation, normalized_geo)
        angle_groups = []

        for geo in polygons:
            normalized, rotation = self._normalize_geometry_with_rotation(geo)
            found = False
            for i, (orig_rep, rot_rep, norm_rep) in enumerate(groups):
                if self._are_geometries_equivalent(
                    normalized, norm_rep, tolerance
                ):
                    rel_angle = (rot_rep - rotation) % 360.0
                    angle_groups[i].append(rel_angle)
                    found = True
                    break
            if not found:
                groups.append((geo, rotation, normalized))
                angle_groups.append([0.0])

        unique_geometries = [orig_rep for orig_rep, _, _ in groups]

        # Generate Image for MarkerAligner
        self._save_geometry_as_png(unique_geometries[0], output_file=file_path)

        return unique_geometries, angle_groups[0]  # TODO: Fix this maybe?

    def _calculate_bounds(self, geometry):
        """
        Calculate the bounding box of a Shapely Polygon or MultiPolygon.
        """
        if isinstance(geometry, MultiPolygon):
            # Get bounds for all polygons in the MultiPolygon
            bounds = [polygon.bounds for polygon in geometry.geoms]
            min_x = min(b[0] for b in bounds)
            min_y = min(b[1] for b in bounds)
            max_x = max(b[2] for b in bounds)
            max_y = max(b[3] for b in bounds)
            return min_x, min_y, max_x, max_y
        elif isinstance(geometry, Polygon):
            # Get bounds for a single Polygon
            return geometry.bounds
        else:
            raise ValueError(
                "Unsupported geometry type. Expected Polygon or MultiPolygon."
            )

    def _rescale_coords(self, coords, min_x, min_y, scaling_factor):
        """
        Rescale coordinates based on a scaling factor.
        """
        return [
            ((x - min_x) * scaling_factor, (y - min_y) * scaling_factor)
            for x, y in coords
        ]

    def _draw_polygon(
        self, draw, polygon, min_x, min_y, scaling_factor, fill_color
    ):
        """
        Draw a rescaled polygon (with holes) on an image.
        """
        # Rescale and draw the exterior
        rescaled_exterior = self._rescale_coords(
            polygon.exterior.coords, min_x, min_y, scaling_factor
        )
        draw.polygon(rescaled_exterior, fill=fill_color)

        # Rescale and draw the holes (interiors)
        for interior in polygon.interiors:
            rescaled_interior = self._rescale_coords(
                interior.coords, min_x, min_y, scaling_factor
            )
            draw.polygon(rescaled_interior, fill="white")

    def _save_geometry_as_png(
        self,
        geometry,
        target_resolution=600,
        output_file="output.png",
        fill_color="black",
    ):
        """
        Save a Shapely Polygon or MultiPolygon as a PNG image.
        """
        # Calculate the bounds of the geometry
        min_x, min_y, max_x, max_y = self._calculate_bounds(geometry)

        # Calculate the width and height of the bounding box
        width = max_x - min_x
        height = max_y - min_y

        # Determine the scaling factor to fit the geometry into the target resolution
        scaling_factor = min(
            target_resolution / width, target_resolution / height
        )

        # Calculate the new image size based on the scaling factor
        new_width = int(width * scaling_factor)
        new_height = int(height * scaling_factor)

        # Create a blank image with a white background
        image = PIL.Image.new("RGB", (new_width, new_height), "white")
        draw = PIL.ImageDraw.Draw(image)

        # Draw each polygon in the MultiPolygon (or the single Polygon)
        if isinstance(geometry, MultiPolygon):
            for polygon in geometry.geoms:
                self._draw_polygon(
                    draw, polygon, min_x, min_y, scaling_factor, fill_color
                )
        else:
            self._draw_polygon(
                draw, geometry, min_x, min_y, scaling_factor, fill_color
            )

        # Save the image as a PNG file
        image.save(output_file)
        print(f"Image saved as {output_file}")

    def get_cell_by_name(self, cell_name: str) -> pya.Cell:
        """Retrieve a cell by its name from the GDS layout.

        Args:
            cell_name: Name of the cell to retrieve. Case-sensitive.

        Returns:
            pya.Cell: The requested cell object.

        Raises:
            TypeError: If input is not a string
            KeyError: If no cell with specified name exists
        """
        # Input validation
        if not isinstance(cell_name, str):
            raise TypeError(
                f"Expected string for cell name, got {type(cell_name)}"
            )

        # Efficient search using layout's cell dictionary
        cell = self.layout.cell(cell_name)
        if cell is None:
            available_cells = [c.name for c in self.layout.each_cell()]
            raise KeyError(
                f"Cell '{cell_name}' not found in GDS layout. "
                f"Available cells: {', '.join(available_cells[:5])}..."
            )
        return cell

    def _merged_polygons_and_their_positions(self, child_cell, layer, z_pos):

        polygons = self._gather_polygons_in_child_cell(child_cell, layer)
        shapely_polygons = self._polygons_to_shapely(polygons)
        merged_polygons = self._merge_touching_polygons(shapely_polygons)

        positions = [
            [m_pol.centroid.x, m_pol.centroid.y, z_pos]
            for m_pol in merged_polygons
        ]
        return merged_polygons, positions

    def get_marker_aligner(
        self,
        cell_name: str,
        project: Optional[Project] = None,
        marker_layer: Tuple[int, int] = (254, 254),
        marker_height: float = 0.33,
        image_resource: Optional[Image] = None,
        **marker_aligner_kwargs: Dict,
    ) -> MarkerAligner:
        """Create and configure a MarkerAligner from GDS markers.

        Args:
            cell_name: Name of the cell containing markers
            project: Optional Project for resource management
            marker_layer: Layer/datatype tuple for marker identification
            marker_height: Z-height for marker polygons
            image_resource: Optional pre-configured Image resource
            **marker_aligner_kwargs: Additional MarkerAligner configuration

        Returns:
            Configured MarkerAligner instance

        Raises:
            ValueError: If no markers found or invalid input dimensions
            TypeError: For invalid input types
            RuntimeError: If image processing fails
        """
        # Input validation
        if not isinstance(marker_layer, tuple) or len(marker_layer) != 2:
            raise TypeError("marker_layer must be a (int, int) tuple")
        if marker_height < 0:
            raise ValueError("marker_height must be non-negative")

        try:
            cell = self.get_cell_by_name(cell_name)
        except KeyError as e:
            raise ValueError(f"Cell '{cell_name}' not found in layout") from e

        # Polygon processing
        marker_polygons, marker_positions = (
            self._merged_polygons_and_their_positions(
                cell, marker_layer, marker_height
            )
        )

        if not marker_polygons:
            raise ValueError(f"No markers found on layer {marker_layer}")
        if len(marker_positions) < 3:
            raise ValueError("At least 3 markers required for alignment")

        # Image resource handling
        image_dir = f"./images_{self.gds_name}_{marker_layer}"
        self._ensure_folder_exist_else_create(image_dir)

        image_file_path = os.path.join(image_dir, f"marker_{marker_layer}.png")
        _image = image_resource or Image(
            name=str(marker_layer), file_path=image_file_path
        )

        if project is not None:
            if not isinstance(project, Project):
                raise TypeError("project must be a Project instance")
            project.load_resources(_image)

        # Marker processing
        _, marker_orientations = (
            self._group_equivalent_polygons_and_output_image(
                marker_polygons, file_path=image_file_path
            )
        )

        try:
            marker_size = [
                marker_polygons[0].bounds[2] - marker_polygons[0].bounds[0],
                marker_polygons[0].bounds[3] - marker_polygons[0].bounds[1],
            ]
        except:
            UserWarning(
                "Failed to calculate marker sizes based on GDS-polygons."
                " Default [5.0,5.0] will be used instead."
            )
            marker_size = [5.0, 5.0]

        if "max_outliers" not in marker_aligner_kwargs:
            marker_aligner_kwargs["max_outliers"] = (
                len(marker_positions) - 3 if len(marker_positions) >= 3 else 0
            )
        marker_aligner = MarkerAligner(
            name=f"{marker_layer}",
            image=_image,
            marker_size=marker_size,
            **marker_aligner_kwargs,
        )

        marker_aligner.set_markers_at(
            positions=marker_positions,
            orientations=marker_orientations,
        )

        return marker_aligner

    def get_coarse_aligner(
        self,
        cell_name: str,
        coarse_layer: Tuple[int, int] = (200, 200),
        residual_threshold: float = 10.0,
    ) -> CoarseAligner:
        """Create a CoarseAligner from anchor points in GDS.

        Args:
            cell_name: Name of the cell containing coarse alignment features
            coarse_layer: Layer/datatype tuple for anchor identification
            residual_threshold: Maximum allowed alignment residual

        Returns:
            Configured CoarseAligner instance

        Raises:
            ValueError: If no anchors found or invalid threshold
        """
        if not isinstance(coarse_layer, tuple) or len(coarse_layer) != 2:
            raise TypeError("marker_layer must be a (int, int) tuple")
        if residual_threshold <= 0:
            raise ValueError("residual_threshold must be positive")

        cell = self.get_cell_by_name(cell_name)
        _, anchor_positions = self._merged_polygons_and_their_positions(
            cell, coarse_layer, 0
        )

        return CoarseAligner(
            name=f"{cell.name}{coarse_layer}",
            residual_threshold=residual_threshold,
        ).set_coarse_anchors_at(anchor_positions)

    def get_custom_interface_aligner(
        self,
        cell_name: str,
        interface_layer: Tuple[int, int] = (255, 255),
        scan_area_sizes: Optional[List[List[float]]] = None,
        **interface_aligner_kwargs: Dict,
    ) -> InterfaceAligner:
        """Create an InterfaceAligner with custom scan areas from GDS.

        Args:
            cell_name: Name of the cell containing interface features
            interface_layer: Layer/datatype tuple for scan areas
            scan_area_sizes: Optional list of [width, height] pairs
            **interface_aligner_kwargs: Additional InterfaceAligner config

        Returns:
            Configured InterfaceAligner instance
        """
        if not isinstance(interface_layer, tuple) or len(interface_layer) != 2:
            raise TypeError("marker_layer must be a (int, int) tuple")

        cell = self.get_cell_by_name(cell_name)
        scan_area_sizes_polygons, anchor_positions = (
            self._merged_polygons_and_their_positions(cell, interface_layer, 0)
        )

        scan_area_sizes = (
            [
                [
                    scan_area_sizes_polygons[i].bounds[2]
                    - scan_area_sizes_polygons[i].bounds[0],
                    scan_area_sizes_polygons[i].bounds[3]
                    - scan_area_sizes_polygons[i].bounds[1],
                ]
                for i in range(len(scan_area_sizes_polygons))
            ]
            if scan_area_sizes is None
            else scan_area_sizes
        )

        return InterfaceAligner(
            name=f"{cell.name}{interface_layer}",
            **interface_aligner_kwargs,
        ).set_interface_anchors_at(
            positions=anchor_positions,
            scan_area_sizes=scan_area_sizes,
        )



import unittest
from unittest.mock import patch
from npxpy.nodes.project import Project
from npxpy.resources import Image, Mesh
from npxpy.preset import Preset

# Define test resource paths
TEST_IMAGE_PATH = "test_resources/78eab7abd2cd201630ba30ed5a7ef4fc/markers.png"
TEST_MESH_PATH = (
    "test_resources/5416ba193f0bacf1e37be08d5c249914/combined_file.stl"
)


class TestProjectClass(unittest.TestCase):
    def setUp(self):
        # Set up dummy data for presets, image, and mesh
        self.preset_1 = Preset(
            name="Test Preset 1",
            valid_objectives=["25x"],
            valid_resins=["IP-n162"],
        )
        self.preset_2 = Preset(
            name="Test Preset 2",
            valid_objectives=["25x"],
            valid_resins=["IP-Visio"],
        )

        self.image = Image(file_path=TEST_IMAGE_PATH)
        self.mesh = Mesh(file_path=TEST_MESH_PATH)

    def test_project_initialization(self):
        # Create a valid project
        with patch("os.getlogin", return_value="test_user"):
            project = Project(objective="25x", resin="IP-n162", substrate="*")

        self.assertEqual(project.objective, "25x")
        self.assertEqual(project.resin, "IP-n162")
        self.assertEqual(project.substrate, "*")
        self.assertEqual(project.project_info["author"], "test_user")
        self.assertTrue("creation_date" in project.project_info)

    def test_invalid_objective(self):
        # Test for invalid objective value
        with self.assertRaises(ValueError):
            Project(objective="100x", resin="IP-n162", substrate="*")

    def test_invalid_resin(self):
        # Test for invalid resin value
        with self.assertRaises(ValueError):
            Project(objective="25x", resin="Unknown Resin", substrate="*")

    def test_invalid_substrate(self):
        # Test for invalid substrate value
        with self.assertRaises(ValueError):
            Project(
                objective="25x", resin="IP-n162", substrate="InvalidSubstrate"
            )

    def test_load_presets(self):
        # Create a valid project and load presets
        project = Project(objective="25x", resin="IP-n162", substrate="*")
        project.load_presets(self.preset_1)
        self.assertIn(self.preset_1, project.presets)

        # Test loading multiple presets
        project.load_presets([self.preset_1, self.preset_2])
        self.assertIn(self.preset_2, project.presets)

    def test_load_invalid_presets(self):
        # Create a valid project and test loading invalid presets
        project = Project(objective="25x", resin="IP-n162", substrate="*")
        with self.assertRaises(TypeError):
            project.load_presets("Invalid Preset")  # Not a Preset instance

    def test_load_image_and_mesh(self):
        # Create a valid project and load image and mesh resources
        project = Project(objective="25x", resin="IP-n162", substrate="*")
        project.load_resources(self.image)
        project.load_resources(self.mesh)

        self.assertIn(self.image, project.resources)
        self.assertIn(self.mesh, project.resources)

    def test_load_invalid_resources(self):
        # Create a valid project and test loading invalid resources
        project = Project(objective="25x", resin="IP-n162", substrate="*")
        with self.assertRaises(TypeError):
            project.load_resources(
                "Invalid Resource"
            )  # Not an Image or Mesh instance

    @patch("os.path.isfile", return_value=True)
    def test_nano_file_creation(self, mock_isfile):
        # Test the .nano file creation
        project = Project(objective="25x", resin="IP-n162", substrate="*")
        project.load_presets(self.preset_1)
        project.load_resources(self.image)

        # Mock resource file path and name for adding to zip
        with patch(
            "builtins.open", unittest.mock.mock_open(read_data="data")
        ), patch("zipfile.ZipFile.writestr"):
            project.nano(project_name="TestProject", path="./test_path")

        # Check that resource was written to the zip file
        self.assertTrue(mock_isfile.called)


if __name__ == "__main__":
    unittest.main()



import numpy as np
from stl import mesh
import pyvista as pv
from copy import deepcopy


def generate_menger_sponge_positions(order=3, smallest_cube_size=20.0):
    """
    Generates the positions for all cubes in a Menger sponge of a given order.

    Parameters:
    - order (int): The order of the Menger sponge (default is 3).
    - smallest_cube_size (float): The size of the smallest cube (voxel), default is 20.0 units.

    Returns:
    - List of tuples: Each tuple represents the (x, y, z) origin of a cube.
    """
    divisions = 3**order  # Total number of divisions per axis
    positions = []

    for x in range(divisions):
        for y in range(divisions):
            for z in range(divisions):
                cx, cy, cz = x, y, z
                is_solid = True
                for _ in range(order):
                    if ((cx % 3 == 1) + (cy % 3 == 1) + (cz % 3 == 1)) >= 2:
                        is_solid = False
                        break
                    cx //= 3
                    cy //= 3
                    cz //= 3
                if is_solid:
                    # Calculate the position by scaling with the smallest cube size
                    pos = (
                        x * smallest_cube_size,
                        y * smallest_cube_size,
                        z * smallest_cube_size,
                    )
                    positions.append(pos)
    return positions


def main(order=3, smallest_cube_size=20.0, stl_filename="voxel_cube.stl"):
    """
    Generates a Menger sponge, loads the voxel STL, and visualizes it using PyVista.

    Parameters:
    - order (int): The order of the Menger sponge (default is 3).
    - smallest_cube_size (float): The size of the smallest cube (voxel), default is 20.0 units.
    - stl_filename (str): Path to the voxel STL file (default is 'voxel_cube.stl').
    """
    # Generate positions for the Menger sponge
    positions = generate_menger_sponge_positions(
        order=order, smallest_cube_size=smallest_cube_size
    )
    print(
        f"Number of cubes in Menger sponge of order {order}: {len(positions)}"
    )

    # Load the voxel cube STL
    try:
        voxel_mesh = mesh.Mesh.from_file(stl_filename)
    except FileNotFoundError:
        print(f"Error: The file '{stl_filename}' was not found.")
        return

    # Initialize a PyVista plotter
    plotter = pv.Plotter()

    # Append all voxel meshes at their respective positions
    for pos in positions:
        # Create a translated copy of the voxel mesh
        translated_mesh = deepcopy(voxel_mesh)
        translated_mesh.vectors += pos
        # Convert numpy-stl mesh to PyVista mesh
        points = translated_mesh.vectors.reshape(-1, 3)
        faces = np.arange(len(points)).reshape(-1, 3)
        faces = np.hstack([np.full((faces.shape[0], 1), 3), faces])

        pv_mesh = pv.PolyData(points, faces)
        plotter.add_mesh(pv_mesh, color="orange", show_edges=False)

    # Set plotter properties for better visualization
    plotter.set_background("white")
    plotter.show_grid()
    plotter.show_axes()

    # Display the plot
    plotter.show()


if __name__ == "__main__":
    # You can change the order here if needed
    main(order=3, smallest_cube_size=20.0, stl_filename="voxel_cube.stl")
